"""
Script to uninstall IDA plugin.
"""

import sys

from .util import get_ida_user_dir


if __name__ == "__main__":
    ida_dir = get_ida_user_dir()
    if not ida_dir:
        print("Unable to find IDA user directory.")
        sys.exit(1)
    entry_point = ida_dir / "plugins" / "rugosa.py"
    if entry_point.exists():
        entry_point.unlink()
        print(f"Removed {entry_point}")
    else:
        print("Already removed.")



"""
This module uses Dragodis's Flowchart object in order to calculate code paths.
"""
from __future__ import annotations
import functools
from copy import deepcopy

import logging
from typing import TYPE_CHECKING, Optional, Iterable

from dragodis import NotExistError
from dragodis.interface import Flowchart, BasicBlock

if TYPE_CHECKING:
    from .cpu_context import ProcessorContext


logger = logging.getLogger(__name__)


class PathNode:
    """
    Represents a linked-list of objects constituting a path from a specific node to the function entry point node.
    This object can also track cpu context up to a certain address.
    """

    def __init__(self, block: BasicBlock, prev: Optional[PathNode]):
        """
        Initialize a path node.

        :param block: The underlying basic block for this node.
        :param prev: The parent node that points to this node.
        """
        self.block = block
        self.prev = prev
        self._context = None
        self._context_address = None  # address that the context has been filled to (but not including)
        self._init_context = None     # the context used at the starting path
        self._call_depth = 0          # the number of calls deep we are allowed to emulated

    @classmethod
    def iter_all(cls, block: BasicBlock, _visited=None, _cache=None) -> Iterable[PathNode]:
        """
        Iterates all tail path nodes from a given block.

        :param block: Block to obtain all path nodes.
        :param _visited: Internally used.
        :param _cache: Internally used.

        :yields: PathNode objects that represent the last entry of the path linked list.
        """
        if _cache is None:
            _cache = {}

        if _visited is None:
            _visited = set()

        # Otherwise generate path nodes and cache results for next time.
        _visited.add(block.start)

        parents = list(block.blocks_to)
        if not parents:
            yield cls(block, prev=None)
        else:
            for parent in parents:
                if parent.start in _visited:
                    continue

                # Create path nodes for each path of parent.
                for parent_path in cls.iter_all(parent, _visited=_visited, _cache=_cache):
                    key = (block, parent_path)
                    try:
                        yield _cache[key]
                    except KeyError:
                        path_node = cls(block, prev=parent_path)
                        _cache[key] = path_node
                        yield path_node

        _visited.remove(block.start)

    def __bool__(self) -> bool:
        # Implementing this to prevent Python from trying to use __len__() to test truthiness.
        # This avoids wasting time recursively checking the truthiness of self.prev
        return bool(self.block)

    def __len__(self):
        return 1 + (len(self.prev) if self.prev else 0)

    @functools.singledispatchmethod
    def __contains__(self, addr):
        return False

    @__contains__.register
    def _(self, addr: int):
        return addr in self.block

    @__contains__.register
    def _(self, block: BasicBlock):
        return any(_block == block for _block in self)

    def __repr__(self):
        return f"Path({' -> '.join(hex(block.start) for block in self)})"

    def __iter__(self) -> Iterable[BasicBlock]:
        """
        Iterates blocks from the root block to the current.
        """
        if self.prev:
            yield from self.prev
        yield self.block

    def __reversed__(self) -> Iterable[BasicBlock]:
        """
        Iterates blocks from the current block up to the root.
        """
        yield self.block
        if self.prev:
            yield from reversed(self.prev)

    def cpu_context(self, addr: int = None, *, call_depth: int = 0, init_context: ProcessorContext) -> ProcessorContext:
        """
        Returns the cpu context filled to (but not including) the specified ea.

        :param int addr: address of interest (defaults to the last ea of the block)
        :param call_depth: Number of function calls we are allowed to emulate into.
            When we hit our limit (depth is 0), emulation will no longer jump into function calls.
            (Defaults to not emulating into any function calls.)
            NOTE: This does not affect call hooks.
        :param init_context: Initial context to use for the start of the path. (required)

        :return cpu_context.ProcessorContext: cpu context
        """
        logger.debug("Getting context for block starting at: 0x%08X", self.block.start)
        if addr is not None and addr not in self.block:
            raise KeyError(
                f"Provided address 0x{addr:X} not in this block "
                f"(0x{self.block.start:X} :: 0x{self.block.end:X})"
            )

        # Determine address to stop computing.
        if addr is None:
            end = self.block.end  # end of a BasicBlock is the first address after the last instruction.
        else:
            end = addr

        # Determine if we need to force the creation of a new context if we have a different init_context
        # or call_depth.
        new_init_context = self._init_context != init_context or self._call_depth != call_depth
        self._init_context = init_context
        self._call_depth = call_depth

        assert end is not None
        # Fill context up to requested endpoint.
        if self._context_address != end or new_init_context:
            # Create context if:
            #   - not created
            #   - current context goes past requested ea
            #   - given init_context/call_depth is different from the previously given init_context/call_depth.
            if not self._context or self._context_address > end or new_init_context:
                # Need to check if there is a prev, if not, then we need to create a default context here...
                if self.prev:
                    self._context = self.prev.cpu_context(call_depth=call_depth, init_context=init_context)
                    # Modify the context for the current branch if required
                    self._context.prep_for_branch(self.block.start)
                else:
                    self._context = deepcopy(init_context)

                self._context_address = self.block.start

            if self._context_address != end:
                # Fill context up to requested ea.
                logger.debug("Emulating instructions 0x%08X -> 0x%08X", self._context_address, end)
                for line in self.block.lines(start=self._context_address):
                    if line.address == end:
                        break
                    self._context.execute(line.address, call_depth=call_depth)

            self._context_address = end

        # Set the next instruction pointer to be the end instruction that we did NOT execute.
        self._context.ip = end

        return deepcopy(self._context)


def iter_paths(flowchart: Flowchart, addr: int, _cache=None) -> Iterable[PathNode]:
    """
    Given an EA, iterate over the paths to the EA.

    For usage example, see Emulator.iter_context_at()

    ..warning:: DO NOT WRAP THIS GENERATOR IN list()!!!  This generator will iterate all possible paths to the node containing
    the specified EA.  On functions containing large numbers of jumps, the number of paths grows exponentially and
    you WILL hit memory exhaustion limits, extremely slow run times, etc. Use extremely conservative constraints
    when iterating.  Nodes containing up to at least 32,768 paths are computed in a reasonably sane amount of time,
    though it probably doesn't make much sense to check this many paths for the data you are looking for.

    :param flowchart: Dragodis Flowchart to get basic blocks from.
    :param addr: Address of interest
    :param _cache: Internally used.

    :yield: a path to the object
    """
    if _cache is None:
        _cache = {}

    # Obtain the block containing the address of interest
    try:
        block = flowchart.get_block(addr)
    except NotExistError:
        # If block not found, then there are no paths to it.
        logger.debug(f"Unable to find block with ea: 0x{addr:08X}")
        return

    yield from PathNode.iter_all(block, _cache=_cache)



"""
Utilities for working with encoded/encrypted strings.
"""
from __future__ import annotations
import logging
import re
from string import printable as printable_chars
import sys
from typing import Iterable, Tuple, Union

import dragodis
from rugosa.emulation import Emulator


logger = logging.getLogger(__name__)

# fmt: off
# Codecs used to detect encoding of strings.
CODE_PAGES = [
    "ascii",
    "utf-32-be", "utf-32-le", "utf-16-be", "utf-16-le", "utf-8",  # General (utf-7 omitted)
    "gb18030", "gbk",  # Unified Chinese
    "gb2312", "hz",  # Simplified Chinese
    "big5hkscs", "big5",  # Traditional Chinese (cp950 omitted)
    "koi8-r", "iso8859-5", "cp1251", "mac-cyrillic",  # Cyrillic (cp866, cp855 omitted)
    "cp949",  # Korean (johab, iso2022-kr omitted)
    "iso8859-6", "cp1256",  # Arabic (cp864, cp720 omitted)
    "latin1",  # If all else fails, latin1 is always successful.
]
# fmt: on

# CODEC to use for displaying strings in IDA, etc.
DISPLAY_CODE = "cp437" if sys.platform == "win32" else "ascii"


# TODO: Add support for forcing only utf-16 or utf-8 string data?
def find_string_data(data: bytes) -> Iterable[Tuple[int, bytes, str]]:
    """
    Iterates string data found within the given data.
    This looks for the start of UTF-8 or UTF-16 strings found within the given block of data.

    :param data: Chunk of data that contains a mixture of UTF-8 and UTF-16 encoded strings.
    :yields: (start_offset, string_data, encoding)
    """
    if not data or data in (b"\x00", b"\x00\x00"):
        return

    # Iterate strings split up by either \x00\x00 or \x00.
    # If we find a single character followed by a \x00, this is most likely a character of utf-16.
    offset = 0
    buffer = b""
    utf_16 = False
    for entry in re.split(b"((?<=\x00[^\x00]\x00)\x00\x00|\x00)", data):
        buffer += entry
        offset += len(entry)
        string_offset = offset - len(buffer)

        # If we have a single character, then this is most likely part of a utf-16 string.
        # Don't yield anything yet.
        if len(entry) == 1 and entry != b"\x00":
            utf_16 = True

        # If we find a single null character and not currently extracting a utf-16 string,
        # we found the end of a utf-8 string.
        elif entry == b"\x00" and not utf_16:
            if buffer not in (b"\x00", b"\x00\x00"):
                yield string_offset, buffer, "utf-8"
            buffer = b""

        # If we find a double null character, this is the end of a utf-16 string.
        elif entry == b"\x00\x00":
            if buffer not in (b"\x00", b"\x00\x00"):
                yield string_offset, buffer, "utf-16-le"
            buffer = b""
            utf_16 = False

        # If we find a multi-character string but in utf_16 mode, then we misidentified the utf-16.
        # Yield buffer as single character utf-8 strings and then reset buffer to entry.
        elif len(entry) > 1 and utf_16:
            split_strings = []
            for _entry in re.split(b"(\x00)", buffer[:-len(entry)]):
                if _entry == b"\x00":
                    split_strings[-1] += _entry
                else:
                    split_strings.append(_entry)
            for _entry in split_strings:
                if _entry:
                    yield string_offset, _entry, "utf-8"
                string_offset += len(_entry)
            buffer = buffer[-len(entry):]
            utf_16 = False

    # Yield remaining string that may not have a null terminator.
    if buffer and buffer not in (b"\x00", b"\x00\x00"):
        string_offset = offset - len(buffer)
        if utf_16:
            # If we have no ending null byte and there is a null byte in front, then it must be big endian.
            if not buffer.endswith(b"\x00") and len(buffer) < len(data) and data[string_offset-1] == 0x00:
                buffer = b"\x00" + buffer
                yield string_offset - 1, buffer, "utf-16-be"
            else:
                yield string_offset, buffer, "utf-16-le"
        else:
            yield string_offset, buffer, "utf-8"


def get_terminated_bytes(dis: dragodis.Disassembler, addr: int, unit_width: int = 1) -> bytes:
    """
    Extracts null terminated bytes from given address.

    NOTE: This is different from dis.get_string_bytes() since it has no requirement for the data
    range to be a valid string. It just purely gets bytes up to the first null terminator.

    :param dis: Dragodis disassembler
    :param addr: Starting address to pull bytes from.
    :param unit_width: Byte width of string character.
    :return: Null terminated bytes.
    """
    terminator_address = dis.find_bytes(b"\x00" * unit_width, start=addr)
    if terminator_address == -1:
        raise ValueError(f"Unable to locate null terminator for 0x{addr:0X}")
    return dis.get_bytes(addr, terminator_address - addr)


_api_names = [
    ("GetModuleHandleA", 0),
    ("GetModuleHandleW", 0),
    ("LoadLibraryA", 0),
    ("LoadLibraryW", 0),
    ("GetProcAddress", 1),
]


def find_api_resolve_strings(dis: dragodis.Disassembler) -> Iterable[Tuple[int, str]]:
    """
    Finds strings used in API resolution functions (e.g. GetProcAddress)

    :param dis: Dragodis disassembler

    :yields: (address, string) for API string.
    """
    seen = set()
    emulator = Emulator(dis)
    for api_name, arg_index in _api_names:
        try:
            imp = dis.get_import(api_name)
        except dragodis.NotExistError:
            continue
        for address in imp.calls_to:
            try:
                ctx = emulator.context_at(address)
                args = ctx.function_args
            except dragodis.NotExistError as e:
                logger.warning(f"Failed to emulate at 0x{address:08x}: {e}")
                continue

            if len(args) <= arg_index:
                continue

            ptr = args[arg_index].value
            if ptr in seen:
                continue

            # Only include strings actually found in sample statically.
            if not dis.is_loaded(ptr):
                continue

            try:
                string = ctx.memory.read_string(ptr, wide=api_name.endswith("W"))
                yield ptr, string
                seen.add(ptr)
            except UnicodeDecodeError:
                continue


def is_library_string(dis: dragodis.Disassembler, address: int) -> bool:
    """
    Attempts to determine whether the string at the given address is only used in library functions.

    :param dis: Dragodis disassembler
    :param address: Address pointing to string.
    :return:
    """
    found_function = False
    for ref in dis.references_to(address):
        try:
            func = dis.get_function(ref.from_address)
        except dragodis.NotExistError:
            continue
        found_function = True
        if not func.is_library:
            return False
    return found_function


def is_code_string(dis: dragodis.Disassembler, address: int, *, code_segment=None):
    """
    Determines whether the string has a reference to an instruction in the code segment.

    :param dis: Dragodis disassembler
    :param address: Address of the string
    :param code_segment: Segment containing instruction code.
        (Determined using entry point if not provided)
    :return:
    """
    if not code_segment:
        code_segment = dis.get_segment(dis.entry_point)
    return any(ref.from_address in code_segment for ref in dis.references_to(address))


def find_user_strings(
        dis: dragodis.Disassembler, min_length=3, ignore_api=True, ignore_library=True, printable=True, unique=False,
        in_code=True,
) -> Iterable[Tuple[int, str]]:
    """
    Finds user strings that are used within the code segment.

    :param dis: Dragodis disassembler
    :param min_length: The minimum length to count as a string.
    :param ignore_api: Whether to attempt to ignore strings used for API resolution (e.g. GetProcAddress parameters)
        NOTE: This can be slow. Disable this option if performance is a concern.
    :param ignore_library: Whether to ignore strings only used in library functions.
    :param printable: Whether to only include strings printable as ASCII.
    :param unique: Whether to only include the first instance of a string.
        (ie. ignore the same string just with a different addresses)
    :param in_code: Whether to only include strings referenced in the main user code.

    :yields: (address, string)
    """
    seen = set()
    code_segment = dis.get_segment(dis.entry_point)
    api_strings = None

    for entry in dis.strings(min_length):
        string = str(entry)

        if unique:
            if string in seen:
                continue
            seen.add(string)

        # NOTE: Using string.printable set over str.isprintable() since the latter doesn't count whitespace characters like \n
        if printable and not all(c in printable_chars for c in string):
            continue

        if in_code and not is_code_string(dis, entry.address, code_segment=code_segment):
            continue

        if ignore_library and is_library_string(dis, entry.address):
            continue

        if ignore_api:
            if api_strings is None:
                api_strings = list(find_api_resolve_strings(dis))
            if any(address == entry.address for address, _ in api_strings):
                continue

        yield entry.address, string


def _num_raw_bytes(string: str) -> int:
    """
    Returns the number of raw bytes found in the given unicode string
    """
    count = 0
    for char in string:
        char = char.encode("unicode-escape")
        count += char.startswith(b"\\x") + char.startswith(b"\\u") * 2
    return count


def detect_encoding(data: bytes, code_pages=None) -> str:
    """
    Detects the best guess string encoding for the given data.
    NOTE: This will default to "latin1" as a fallback.

    :param data: Data to detect encoding
    :param code_pages: List of possible codecs to try.
        There is a default, but feel free to provide your own.

    :returns: Decoded string and encoding used.
    """
    if code_pages is None:
        code_pages = CODE_PAGES
    best_score = len(data)  # lowest score is best
    best_code_page = None
    best_output = None
    for code_page in code_pages:
        try:
            output = data.decode(code_page).rstrip(u"\x00")
        except UnicodeDecodeError:
            # If it's UTF we may need to strip away some null characters before decoding.
            if code_page in ("utf-16-le", "utf-16-be", "utf-32-le", "utf-32-be"):
                data_copy = data
                while data_copy and data_copy[-1] == 0:
                    try:
                        data_copy = data_copy[:-1]
                        output = data_copy.decode(code_page).rstrip(u"\x00")
                    except UnicodeDecodeError:
                        continue
                    break  # successfully decoded
                else:
                    continue
            # otherwise the code page isn't correct.
            else:
                continue

        score = _num_raw_bytes(output)
        if not best_output or score < best_score:
            best_score = score
            best_output = output
            best_code_page = code_page

    if best_output:
        return best_code_page

    # We shouldn't hit here since "latin1" should at least hit, but just incase...
    return "unicode_escape"


def force_to_string(data: bytes, code_pages=None) -> str:
    """
    Forces given bytes into a string using best guess encoding.

    :param data: Bytes to convert to string.
    :param code_pages: List of possible codecs to try.
        There is a default, but feel free to provide your own.

    :return: Decoded string.
    """
    if code_pages is None:
        code_pages = CODE_PAGES
    try:
        return data.decode(detect_encoding(data, code_pages=code_pages))
    except UnicodeDecodeError:
        return data.decode("latin1")


class DecodedString:
    """
    Holds information about a decoded/decrypted string.

    :param dec_data: Decrypted/decoded string data
    :param enc_data: Original encrypted/encoded string data
    :param encoding: Known encoding used to decoded data into a string
        If not provided, this will be detected using dec_data.
    :param enc_source: The address or dragodis variable object where the enc_data was found.
    :param dec_source: The address or dargodis variable object where the dec_data was found.
    """

    _MAX_COMMENT_LENGTH = 130
    _MAX_NAME_LENGTH = 30

    def __init__(
            self,
            dec_data: bytes,
            enc_data: bytes = None,
            encoding: str = None,
            enc_source: Union[None, int, dragodis.interface.Variable] = None,
            dec_source: Union[None, int, dragodis.interface.Variable] = None,
    ):
        self.data = dec_data
        self.enc_data = enc_data
        self.encoding = encoding or detect_encoding(dec_data)
        self.enc_source = enc_source
        self.dec_source = dec_source

    def __str__(self):
        """
        Detects and decodes string data.
        """
        return self.data.decode(self.encoding).rstrip("\x00")

    def __bytes__(self):
        return self.data

    @property
    def display_name(self) -> str:
        """Returns a disassembler friendly, printable name for the decoded string."""
        return str(self).encode(DISPLAY_CODE, "replace").decode(DISPLAY_CODE)

    def _annotate(
            self,
            dis: dragodis.Disassembler,
            item: Union[int, dragodis.interface.Line, dragodis.interface.Variable],
            name: str = None, comment: str = None
    ):
        """
        Annotates a Dragodis object
        """
        if isinstance(item, int):
            item = dis.get_line(item)
        if name:
            item.name = name
        if comment:
            if isinstance(item, dragodis.interface.GlobalVariable):
                item = dis.get_line(item.address)
                item.set_comment(comment, dragodis.CommentType.repeatable)
            if isinstance(item, dragodis.interface.Line):
                item.set_comment(comment, dragodis.CommentType.repeatable)

    def rename(self, dis: dragodis.Disassembler, name=None):
        """
         Renames (and comments) the string variable in the disassembler.

         :param str name: New name to given encoded string. (defaults to the decoded string itself)
         """
        name = name or self.display_name
        name = name[:self._MAX_NAME_LENGTH]

        # Add comment
        comment = 'Dec: "{}"'.format(name[: self._MAX_COMMENT_LENGTH])
        if len(name) > self._MAX_COMMENT_LENGTH:
            comment += " (truncated)"
        # TODO: Determine what to do when a stack source
        if self.enc_source:
            self._annotate(dis, self.enc_source, name="enc_"+name, comment=comment)
        if self.dec_source:
            self._annotate(dis, self.dec_source, name=name)
            # TODO: Setting datatype on a Variable is not current supported.
            # self.dec_source.data_type = dragodis.DataType.name

    def patch(self, dis: dragodis.Disassembler, fill_char=b"\x00", rename=True):
        """
        Patches the decrypted string data into the underlying disassembler.

        :param dis: Dragodis disassembler
        :param fill_char:
            Character to use to fill left over space if decrypted data
            is shorter than its encrypted data.
            (Set to None to leave the original data)
        :param rename:
            Whether to also rename the variable names.
            (This can also be done manually by using .rename())
        """
        if rename:
            self.rename(dis)

        if self.dec_source:
            dec_data = self.data
            if fill_char:
                dec_data += fill_char * (len(self.enc_data) - len(dec_data))
            dec_source = self.dec_source
            if isinstance(dec_source, int):
                dec_source = dis.get_line(dec_source)
                dec_source.data = dec_data
            elif isinstance(dec_source, dragodis.interface.GlobalVariable):
                dec_source = dis.get_line(dec_source.address)
                dec_source.data = dec_data
            else:
                logger.warning(f"Unable to patch {dec_source!r}")



"""
Functions found in kernel32.dll

Win32 base API
"""

import logging

from . import win_constants as wc
from ...call_hooks import builtin_func
from ... import constants
from ... import actions, objects

logger = logging.getLogger(__name__)


# Maps environment variables to the most common values.
ENV_VAR_MAP = {
    "comspec": "C:\\Windows\\system32\\cmd.exe",
    "windir": "C:\\Windows",
    "systemroot": "C:\\Windows",
    "systemdrive": "C:",
    "homedrive": "C:",
}


@builtin_func("GetEnvironmentVariableA")
@builtin_func("GetEnvironmentVariableW")
#typespec("DWORD GetEnvironmentVariableA(LPCSTR lpName, LPSTR lpBuffer, DWORD nSize)")
def get_environment_variable(cpu_context, func_name, func_args):
    """
    Retrieves the contents of the specified variable from the environment block of the calling process.
    """
    wide = func_name.endswith("W")
    var_name_ptr, buffer_ptr, max_size = func_args
    var_name = cpu_context.memory.read_string(var_name_ptr, wide=wide)

    # Replace some common environment variables, and add %'s for others.
    # - 1 for null terminator
    name = ENV_VAR_MAP.get(var_name.lower(), f"%{var_name}%")[:max_size - 1]
    cpu_context.memory.write_string(buffer_ptr, name, wide=wide)
    logger.debug("Getting environment variable: %s -> %s", var_name, name)

    return len(name)


@builtin_func("GetCurrentDirectoryA")
@builtin_func("GetCurrentDirectoryW")
#typespec("DWORD GetCurrentDirectory(DWORD  nBufferLength, LPTSTR lpBuffer);")
def get_current_directory(cpu_context, func_name, func_args):
    """
    Retrieves the current working directory

    Using value "." to represent the current working directory for this emulator
    """
    wide = func_name.endswith("W")
    max_size, buffer_ptr = func_args

    cwd = u"."[:max_size]

    logger.debug("Writing current working directory %s to 0x%08X", cwd, buffer_ptr)
    cpu_context.memory.write_string(buffer_ptr, cwd, wide=wide)

    return len(cwd)


@builtin_func("GetModuleFileNameA")
@builtin_func("GetModuleFileNameW")
#typespec("DWORD GetModuleFileNameA_1(HMODULE hModule, LPSTR lpFilename, DWORD nSize);")
def get_module_file_name(cpu_context, func_name, func_args):
    r"""
    Get the fully qualified path for the file that contains the specified module.

    Using the real filename prefixed by "%INPUT_FILE_DIR%\" to indicate the file path for this emulator.
    """
    wide = func_name.endswith("W")
    module_handle, filename_ptr, max_size = func_args

    # We don't support getting filename for modules that aren't itself.
    if module_handle != 0:
        return

    # Getting input file path as module path.
    # Since we shouldn't expose the user's real file path structure we'll use %INPUT_FILE_DIR% instead.
    # Must be truncated to fit in max_size filename_ptr (-1 for the terminator)
    disassembler = cpu_context.emulator.disassembler
    file_path = f"%INPUT_FILE_DIR%\\{disassembler.input_path.name}"
    file_path = file_path[:max_size - 1]

    logger.debug("Writing module path %s to 0x%08X", file_path, filename_ptr)
    cpu_context.memory.write_string(filename_ptr, file_path, wide=wide)

    return len(file_path)


@builtin_func("GetShortPathNameA")
@builtin_func("GetShortPathNameW")
#typespec("DWORD GetShortPathNameA(LPCSTR lpszLongPath, LPSTR lpszShortPath, DWORD cchBuffer)")
def get_short_path_name(cpu_context, func_name, func_args):
    """
    Retrieves the short path form of the specified path.
    """
    # We aren't going to bother actually shortening the path name.
    # This hook is just here to ensure the destination gets filled in.
    wide = func_name.endswith("W")
    long_path_ptr, short_path_ptr, short_path_size = func_args

    long_path = cpu_context.memory.read_string(long_path_ptr, wide=wide)

    short_path = long_path[:short_path_size]
    cpu_context.memory.write_string(short_path_ptr, short_path, wide=wide)
    logger.debug("Copying %r from 0x%08X -> 0x%08X", short_path, long_path_ptr, short_path_ptr)

    return len(short_path)


@builtin_func("GetSystemDirectoryA")
@builtin_func("GetSystemDirectoryW")
#typespec("UINT GetSystemDirectoryA(char* lpBuffer, UINT uSize);")
#typespec("UINT GetSystemDirectoryW(wchar* lpbuffer, UINT uSize);")
def get_system_directory(cpu_context, func_name, func_args):
    r"""
    Retrieves the path of the system directory.

    Using unexpanded "%WinDir%\System32" to indicate the system directory for this emulator.
    """
    wide = func_name.endswith("W")
    buffer_ptr, max_size = func_args

    system_dir = r"%WinDir%\System32"[:max_size]

    logger.debug("Writing system directory %s to 0x%08X", system_dir, buffer_ptr)
    cpu_context.memory.write_string(buffer_ptr, system_dir, wide=wide)

    return len(system_dir)


@builtin_func("GetTempPathA")
@builtin_func("GetTempPathW")
#typespec("DWORD GetTempPathA(DWORD nBufferLength,LPSTR lpBuffer);")
def get_temp_path(cpu_context, func_name, func_args):
    """
    Retrieves the path of the temp directory

    Using unexpanded "%Temp%\" to indicate the temp directory for this emulator
    """
    wide = func_name.endswith("W")
    max_size, buffer_ptr = func_args

    # The returned string ends with a backslash
    temp_dir = u"%Temp%\\"[:max_size]

    logger.debug("Writing temp directory %s to 0x%08X}", temp_dir, buffer_ptr)
    cpu_context.memory.write_string(buffer_ptr, temp_dir, wide=wide)

    return len(temp_dir)


@builtin_func("GetTickCount")
@builtin_func("GetTickCount64")
#typespec("DWORD GetTickCount();")
#typespec("ULONGLONG GetTickCount64();")
def get_tick_count(cpu_context, func_name, func_args):
    """
    Retrieves the number of milliseconds that have elapsed since the system was started, up to 49.7 days.
    """
    # TODO: Should we allow users to specify a specific tick count or provide permission
    #   to use the real time when setting up Emulator?
    # mask = 0xFFFFFFFFFFFFFFFF if func_name == "GetTickCount64" else 0xFFFFFFFF
    # return int(time.time()) & mask

    # Returning a fake tick count in order for results to be more deterministic.
    return 1587410693


@builtin_func("GetWindowsDirectoryA")
@builtin_func("GetWindowsDirectoryW")
#typespec("UINT GetWindowsDirectoryA(char* lpBuffer, UINT uSize);")
#typespec("UINT GetWindowsDirectoryW(wchar* lpBuffer, UINT uSize);")
def get_windows_directory(cpu_context, func_name, func_args):
    """
    Retrieves the path of the Windows directory.

    Using unexpanded "%WinDir%" to indicate the windows directory for this emulator.
    """
    wide = func_name.endswith("W")
    buffer_ptr, max_size = func_args

    win_dir = "%WinDir%"[:max_size]
    logger.debug("Writing windows directory %s to 0x%08X", win_dir, buffer_ptr)
    cpu_context.memory.write_string(buffer_ptr, win_dir, wide=wide)

    return len(win_dir)


@builtin_func("CreateNamedPipeA")
#typespec("HANDLE CreateNamedPipeA(char* lpName, DWORD dwOpenMode, DWORD dwPipeMode, DWORD nMaxInstances, DWORD nOutBufferSize, DWORD nInBufferSize, DWORD nDefaultTimeOut, SECURITY_ATTRIBUTES* lpSecurityAttributes);")
def create_named_pipe(cpu_context, func_name, func_args):
    """
    Creates an instance of a named pipe and returns a handle for subsequent pipe operations.
    """
    # TODO: Create high level NamedPipe objects to keep track of data stored in pipes.
    #   This should work similar if not the same to the File objects we already keep track of.
    return cpu_context.objects.alloc()


@builtin_func("GetComputerNameA")
@builtin_func("GetComputerNameW")
#typespec("BOOL GetComputerNameA(char* lpBuffer, DWORD* nSize);")
#typespec("BOOL GetComputerNameW(wchar* lpBuffer, DWORD* nSize);")
def get_computer_name(cpu_context, func_name, func_args):
    """
    Retrieves the NetBIOS name of the local computer.
    This name is established at system startup, when the system reads it
    from the registry.

    Using the computer name "RUGOSA_COMP" for this emulator.
    TODO: Should we allow user to provide their own computer name when setting up Emulator?
    """
    wide = func_name.endswith("W")
    buffer_ptr, size_ptr = func_args

    computer_name = "RUGOSA_COMP"
    logger.debug("Writing computer name %s to 0x%08X", computer_name, buffer_ptr)
    cpu_context.memory.write_string(buffer_ptr, computer_name, wide=wide)
    cpu_context.memory.write_data(size_ptr, len(computer_name), data_type=constants.DWORD)

    return 1  # return success


@builtin_func("GetUserNameA")
@builtin_func("GetUserNameW")
#typespec("BOOL GetUserNameA(char* lpBuffer, DWORD* pcbBuffer);")
#typespec("BOOL GetUserNameW(wchar* lpBuffer, DWORD* pcbBuffer);")
def get_user_name(cpu_context, func_name, func_args):
    """
    Retrieves the name of the user associated with the current thread.

    Using the user name "rugosa" for this emulator.
    TODO: Should we allow user to provide their own user name when setting up Emulator?
    """
    wide = func_name.endswith("W")
    buffer_ptr, size_ptr = func_args

    user_name = "rugosa"
    logger.debug("Writing user name %s to 0x%08X", user_name, buffer_ptr)
    cpu_context.memory.write_string(buffer_ptr, user_name, wide=wide)
    cpu_context.memory.write_data(size_ptr, len(user_name), data_type=constants.DWORD)

    return 1  # return success


@builtin_func("CreateProcessA")
@builtin_func("CreateProcessW")
#typespec("BOOL CreateProcessA(LPCSTR lpApplicationName, LPSTR lpCommandLine, LPSECURITY_ATTRIBUTES lpProcessAttributes, LPSECURITY_ATTRIBUTES lpThreadAttributes, BOOL bInheritHandles, DWORD dwCreationFlags, LPVOID lpEnvironment, LPCSTR lpCurrentDirectory, LPSTARTUPINFOA lpStartupInfo, LPPROCESS_INFORMATION lpProcessInformation)")
def create_process(cpu_context, func_name, func_args):
    """
    Create a new process.
    """
    wide = func_name.endswith("W")

    app_ptr = func_args[0]
    cmd_ptr = func_args[1]

    cmd = cpu_context.memory.read_string(cmd_ptr, wide=wide)
    if app_ptr:
        app = cpu_context.memory.read_string(app_ptr, wide=wide)
        cmd = app + " " + cmd

    logger.debug("%s: %r", func_name, cmd)
    cpu_context.actions.add(actions.CommandExecuted(cpu_context.ip, cmd))

    return cpu_context.objects.alloc()


@builtin_func("WinExec")
#typespec("UINT WinExec(char* lpCmdLine, UINT nCmdShow);")
def win_exec(cpu_context, func_name, func_args):
    """
    Runs the specified application.
    """
    cmd_ptr, visibility = func_args
    cmd = cpu_context.memory.read_string(cmd_ptr)

    logger.debug("WinExec: %r", cmd)
    cpu_context.actions.add(
        actions.CommandExecuted(cpu_context.ip, cmd, wc.Visibility(visibility))
    )

    return cpu_context.objects.alloc()


@builtin_func("CreateDirectoryA")
@builtin_func("CreateDirectoryW")
#typespec("BOOL CreateDirectoryA(char* lpPathName, SECURITY_ATTRIBUTES* lpSecurityAttributes);")
#typespec("BOOL CreateDirectoryw(wchar* lpPathName, SECURITY_ATTRIBUTES* lpSecurityAttributes);")
def create_directory(cpu_context, func_name, func_args):
    """
    Creates a new directory.
    """
    wide = func_name.endswith("W")
    path_ptr = func_args[0]

    path = cpu_context.memory.read_string(path_ptr, wide=wide)
    logger.debug("Create Directory: %r", path)

    if path:
        cpu_context.actions.add(actions.DirectoryCreated(cpu_context.ip, path))

    return 1  # return success


@builtin_func("CreateFileA")
@builtin_func("CreateFileW")
#typespec("HANDLE CreateFileA(char* lpFileName, DWORD dwDesiredAccess, DWORD dwShareMode, SECURITY_ATTRIBUTES* lpSecurityAttributes, DWORD dwCreationDisposition, DWORD dwFlagsAndAttributes, HANDLE hTemplateFile);")
#typespec("HANDLE CreateFileW(wchar* lpFileName, DWORD dwDesiredAccess, DWORD dwShareMode, SECURITY_ATTRIBUTES* lpSecurityAttributes, DWORD dwCreationDisposition, DWORD dwFlagsAndAttributes, HANDLE hTemplateFile);")
def create_file(cpu_context, func_name, func_args):
    """
    Creates or opens a file or I/O device.
    The function returns a handle that can be used to access the file or device for various
    types of I/O depending on the file or device and the flags and attributes specified.
    """
    wide = func_name.endswith("W")
    name_ptr = func_args[0]

    path = cpu_context.memory.read_string(name_ptr, wide=wide)

    mode = ""
    desired_access = func_args[1]
    if desired_access & wc.GENERIC_READ:
        mode += "r"
    if desired_access & wc.GENERIC_WRITE:
        mode += "w"

    if not path:
        path = f"0x{cpu_context.ip:08x}.bin"

    # Even if we have seen the path before, CreateFile* always creates a new
    # handle, so we should too.
    handle = cpu_context.objects.alloc()

    create_disposition = func_args[4]
    if create_disposition in [wc.CREATE_NEW, wc.CREATE_ALWAYS]:
        logger.debug("Created file: %s", path)
        cpu_context.actions.add(actions.FileCreated(cpu_context.ip, handle, path, mode))
    elif create_disposition in [wc.OPEN_EXISTING, wc.OPEN_ALWAYS]:
        logger.debug("Opened file: %s", path)
        cpu_context.actions.add(actions.FileOpened(cpu_context.ip, handle, path, mode))
    elif create_disposition == wc.TRUNCATE_EXISTING:
        logger.debug("Truncated file: %s", path)
        cpu_context.actions.add(actions.FileTruncated(cpu_context.ip, handle, path, mode))
    
    return handle


@builtin_func("WriteFile")
#typespec("BOOL WriteFile(HANDLE hFile, LPCVOID lpBuffer, DWORD nNumberOfBytesToWrite, LPDWORD lpNumberOfBytesWritten, LPOVERLAPPED lpOverlapped)")
def write_file(cpu_context, func_name, func_args):
    """
    Writes data to the specified file.
    """
    handle = func_args[0]
    data_ptr = func_args[1]
    num_bytes_to_write = func_args[2]
    bytes_written_ptr = func_args[3]

    if not handle:
        logger.warning(f"File with handle {hex(handle)} was never opened.")
        return 1

    logger.debug(f"Writing {num_bytes_to_write} bytes to file with handle {hex(handle)}.")
    data = cpu_context.memory.read(data_ptr, num_bytes_to_write)
    cpu_context.actions.add(actions.FileWritten(cpu_context.ip, handle, data))

    if bytes_written_ptr:
        cpu_context.memory.write_data(bytes_written_ptr, len(data), data_type=constants.DWORD)

    return 1  # return success


@builtin_func("MoveFileA")
@builtin_func("MoveFileExA")
def move_file(cpu_context, func_name, func_args):
    """
    Moves an existing file (or directory) to new location.
    """
    old_name_ptr, new_name_ptr, *_ = func_args
    old_path = cpu_context.memory.read_data(old_name_ptr).decode("utf8")
    new_path = cpu_context.memory.read_data(new_name_ptr).decode("utf8")
    logger.debug("Moving file: %s -> %s", old_path, new_path)

    handle = cpu_context.objects.get_or_alloc(objects.File, path=old_path)

    if not new_path:
        logger.debug("Ignoring attempt to change file path to empty string.")
    
    cpu_context.actions.add(actions.FileMoved(cpu_context.ip, handle, old_path, new_path))

    return 1  # return success


@builtin_func("CloseHandle")
#typespec("BOOL __stdcall CloseHandle(HANDLE hObject)")
def close_file(cpu_context, func_name, func_args):
    """
    Closes an open object handle.
    """
    handle = func_args[0]

    if not handle:
        logger.warning(f"File with handle {hex(handle)} does not exist.")
        return 1
        
    logger.debug(f"Closing file with handle {hex(handle)}.")
    cpu_context.actions.add(actions.FileClosed(cpu_context.ip, handle))

    return 1  # return success


@builtin_func("DeleteFileA")
@builtin_func("DeleteFileW")
#typespec("BOOL DeleteFileA(char* lpFileName);")
#typespec("BOOL DeleteFileW(wchar* lpFileName);")
def delete_file(cpu_context, func_name, func_args):
    """
    Deletes an existing file.
    """
    wide = func_name.endswith("W")
    path_ptr = func_args[0]
    path = cpu_context.memory.read_string(path_ptr, wide=wide)
    logger.debug("Deleting: %s", path)

    handle = cpu_context.objects.get_or_alloc(objects.File, path=path)
    cpu_context.actions.add(actions.FileDeleted(cpu_context.ip, handle, path))

    return 1  # return success


@builtin_func("CreateMutexA")
@builtin_func("CreateMutexW")
@builtin_func("CreateMutexExA")
@builtin_func("CreateMutexExW")
#typespec("HANDLE CreateMutexA(SECURITY_ATTRIBUTES* lpMutexAttributes, BOOL bInitialOwner, char* lpName);")
#typespec("HANDLE CreateMutexW(SECURITY_ATTRIBUTES* lpMutexAttributes, BOOL bInitialOwner, wchar* lpName);")
#typespec("HANDLE CreateMutexExA(SECURITY_ATTRIBUTES* lpMutexAttributes, char* lpName, DWROD dwFlags, DWORD dwDesiredAccess);")
#typespec("HANDLE CreateMutexExW(SECURITY_ATTRIBUTES* lpMutexAttributes, wchar* lpName, DWORD dwFlags, DWORD dwDesiredAccess);")
def create_mutex(cpu_context, func_name, func_args):
    """
    Creates or opens a named or unnamed mutex object and returns a handle to the object.
    """
    return cpu_context.objects.alloc()


@builtin_func("CreateEventA")
@builtin_func("CreateEventW")
@builtin_func("CreateEventExA")
@builtin_func("CreateEventEXW")
#typespec("HANDLE CreateEventA(SECURTIY_ATTRIBUTES* lpEventAttributes, BOOL bManualReset, BOOL bInitialState, char* lpName);")
#typespec("HANDLE CreateEventW(SECURTIY_ATTRIBUTES* lpEventAttributes, BOOL bManualReset, BOOL bInitialState, wchar* lpName);")
#typespec("HANDLE CreateEventExA(SECURITY_ATTRIBUTES* lpEventAttributes, char* lpName, DWORD dwFlags, DWORD dwDesiredAccess);")
#typespec("HANDLE CreateEventExW(SECURITY_ATTRIBUTES* lpEventAttributes, wchar* lpName, DWORD dwFlags, DWORD dwDesiredAccess);")
def create_event(cpu_context, func_name, func_args):
    """
    Creates or opens a named or unnamed event object and returns a handle to the object.
    """
    return cpu_context.objects.alloc()


@builtin_func("CreateSemaphoreA")
@builtin_func("CreateSemaphoreW")
@builtin_func("CreateSemaphoreExA")
@builtin_func("CreateSemaphoreExW")
#typespec("HANDLE CreateSemaphoreA(SECURITY_ATTRIBUTES* lpSemaphoreAttributes, LONG lInitialCount, LONG lMaximumCount, char* lpName);")
#typespec("HANDLE CreateSemaphoreW(SECURITY_ATTRIBUTES* lpSemaphoreAttributes, LONG lInitialCount, LONG lMaximumCount, wchar* lpName);")
#typespec("HANDLE CreateSemaphoreExA(SECURITY_ATTRIBUTES* lpSemaphoreAttributes, LONG lInitialCount, LONG lMaximumCount, char* lpName, DWORD dwFlags, DWORD dwDesiredAccess);")
#typespec("HANDLE CreateSemaphoreExW(SECURITY_ATTRIBUTES* lpSemaphoreAttributes, LONG lInitialCount, LONG lMaximumCount, wchar* lpName, DWORD dwFlags, DWORD dwDesiredAccess);")
def create_semaphore(cpu_context, func_name, func_args):
    """
    Creates or opens a named or unnamed semaphore object and returns a handle to the object.
    """
    return cpu_context.objects.alloc()




from __future__ import annotations

import contextlib
import logging
import traceback
from typing import TYPE_CHECKING, Optional

import jpype

RugosaActionListener = jpype.JClass("dc3.rugosa.plugin.RugosaActionListener")
RugosaTask = jpype.JClass("dc3.rugosa.plugin.RugosaTask")

import dragodis
from dragodis.ghidra import GhidraLocal
from ghidra.util import Swing

import rugosa
from rugosa.ghidra_plugin.gui.mouse_listener import MouseListener
from rugosa.ghidra_plugin.gui.program_listener import ProgramListener
from rugosa.ghidra_plugin.gui.context_menu import TableContextMenuEntry
from rugosa.ghidra_plugin.gui.error_msg import show_exception, catch_errors
from rugosa.ghidra_plugin.components.operands import Operands
from rugosa.ghidra_plugin.components.registers import Registers
from rugosa.ghidra_plugin.components.memory import Memory
from rugosa.ghidra_plugin.components.variables import Variables
from rugosa.ghidra_plugin.components.actions import Actions
from rugosa.ghidra_plugin.components.function_arguments import FunctionArguments
from rugosa.ghidra_plugin.components.call_history import CallHistory
from rugosa.ghidra_plugin.components.stdout import Stdout
from rugosa.emulation.cpu_context import ProcessorContext

if TYPE_CHECKING:
    import ghidra
    from dc3.rugosa.plugin import EmulatorForm as JavaEmulatorForm, RugosaPlugin
    from java.awt.event import ActionEvent, MouseEvent
    from javax.swing import JTable

logger = logging.getLogger(__name__)


class EmulatorForm:

    def __init__(self, plugin: RugosaPlugin):
        self.plugin = plugin
        self.program_listener = ProgramListener(self)
        self.plugin.setProgramListener(self.program_listener)

    @property
    def _ctx(self) -> Optional[ProcessorContext]:
        if state := self.program_listener.state:
            return state.ctx

    @property
    def dis(self) -> GhidraLocal:
        if state := self.program_listener.state:
            return state.dis

    @property
    def emulator(self) -> rugosa.Emulator:
        if state := self.program_listener.state:
            return state.emulator

    @property
    def state(self) -> ProgramListener:
        return self.program_listener.state

    @catch_errors
    def setupUI(self, form: JavaEmulatorForm):
        self.ui = form
        # Remove placeholder text.
        self.ui.instruction.setText("")
        self.ui.status.setText("")
        self.ui.functionSignature.setText("")

        self.operands = Operands(self)
        self.registers = Registers(self)
        self.memory = Memory(self)
        self.variables = Variables(self)
        self.function_arguments = FunctionArguments(self)
        self.actions = Actions(self)
        self.call_history = CallHistory(self)
        self.stdout = Stdout(self)

        # Hook up action listeners.
        self.ui.instruction.addMouseListener(MouseListener(clicked=self.jump_instruction))
        self.ui.runButton.addActionListener(RugosaActionListener @ self.run)
        self.ui.stepOverButton.addActionListener(RugosaActionListener @ self.step_over)
        # Double-clicking table items should try to redirect disassembly if possible.
        redirect = MouseListener(clicked=self.redirect)
        self.ui.operandsTable.addMouseListener(redirect)
        self.ui.registersTable.addMouseListener(redirect)
        self.ui.variablesTable.addMouseListener(redirect)
        self.ui.functionArgumentsTable.addMouseListener(redirect)
        self.ui.callHistoryTable.addMouseListener(redirect)
        self.ui.actionsSideTable.addMouseListener(redirect)
        self.ui.actionsFieldsTable.addMouseListener(redirect)

    def add_context_menu_entry(self, entry: TableContextMenuEntry):
        self.plugin.addContextMenuEntry(entry.NAME, entry)

    @contextlib.contextmanager
    def transaction(self, message="rugosa transaction"):
        """Some modifications with Ghidra require being in a transaction."""
        t = self.dis._program.startTransaction(message)
        success = False
        try:
            yield
            success = True
        finally:
            self.dis._program.endTransaction(t, success)

    @catch_errors
    def jump_instruction(self, event: MouseEvent):
        state = self.program_listener.state
        if state is None:
            return
        if state.ctx and event.getClickCount() == 2:  # double click
            self.plugin.goTo(state.dis._to_addr(state.ctx.ip))

    @catch_errors
    def redirect(self, event: MouseEvent):
        """Redirects clicks on addresses to the line in the disassembly."""
        dis = self.dis
        if dis is None:
            return
        if event.getClickCount() == 2:
            table: JTable = event.getSource()
            row = table.getSelectedRow()
            col = table.getSelectedColumn()
            value = table.getModel().getValueAt(row, col)
            try:
                address = int(value, 16)
            except ValueError:
                return
            self.plugin.goTo(dis._to_addr(address))

    @catch_errors
    def enable_emulator_controls(self):
        """
        Enable emulator buttons based on current context.
        """
        if not self._ctx:
            self.reset()
            return

        self.ui.tabs.setEnabled(True)
        self.ui.stepOverButton.setEnabled(True)
        # TODO: Add support for these after some refactoring of ProcessorContext.
        # self.ui.stepIntoButton.setEnabled(self._ctx.instruction.is_call)
        # self.ui.stepOutButton.setEnabled(bool(self._call_stack))

    @catch_errors
    def update(self):
        """
        Updates UI based on current context.
        """
        ctx = self._ctx
        if not ctx:
            return

        # Enable controls.
        self.enable_emulator_controls()

        insn = ctx.instruction
        self.ui.instruction.setText(f"0x{ctx.ip:08X}: {insn.text}")

        # Fill in tables
        self.operands.populate(ctx)
        self.registers.populate(ctx)
        self.variables.populate(ctx)
        self.function_arguments.populate(ctx)
        self.memory.populate(ctx)
        self.stdout.populate(ctx)
        self.call_history.populate(ctx)
        self.actions.populate(ctx)

        executed_instructions = ctx.executed_instructions
        if executed_instructions:
            self.ui.status.setText(
                f"Emulated {len(executed_instructions)} instructions: "
                f"{hex(ctx.executed_instructions[0])} -> {hex(executed_instructions[-1])}"
            )
        else:
            self.ui.status.setText(f"Emulated 0 instructions.")

        self.plugin.goTo(self.dis._to_addr(ctx.ip))

    def _emulate(self, addr):
        with self.transaction():
            try:
                self.state.ctx = self.emulator.context_at(
                    addr,
                    depth=self.ui.traceDepth.getValue(),
                    call_depth=self.ui.callDepth.getValue(),
                    exhaustive=self.ui.exhaustiveCheckBox.isSelected(),
                    follow_loops=self.ui.followLoopsCheckBox.isSelected(),
                )
                Swing.runIfSwingOrRunLater(self.update)
            except dragodis.NotExistError as e:
                # User hasn't selected a valid function.
                show_exception(e, "Invalid Instruction", level=logging.WARNING)

            except Exception as e:
                show_exception(e, "Failed to emulate")

    @catch_errors
    def run(self, event: ActionEvent):
        """
        Runs emulation using current settings.
        """
        addr = self.plugin.getCurrentAddress().getOffset()
        task = RugosaTask("Emulating...", lambda: self._emulate(addr))
        self.plugin.tool.execute(task)

    @catch_errors
    def step_over(self, event: ActionEvent):
        ctx = self._ctx
        if not ctx:
            return
        ctx.execute(call_depth=self.ui.callDepth.getValue())
        self.update()

    def reset(self):
        self._ctx = None
        self.ui.stepOverButton.setEnabled(False)
        self.ui.stepIntoButton.setEnabled(False)
        self.ui.stepOutButton.setEnabled(False)
        self.ui.tabs.setEnabled(False)


def initialize(form: JavaEmulatorForm, plugin: RugosaPlugin):
    """
    Called by Java to initialize Emulator plugin.
    """
    logging.basicConfig(level=logging.DEBUG)
    try:
        _form = EmulatorForm(plugin)
        Swing.runIfSwingOrRunLater(lambda: _form.setupUI(form))
    except Exception as e:
        logging.exception(e)




from .actions import Actions
from .call_history import CallHistory
from .function_arguments import FunctionArguments
from .memory import Memory
from .operands import Operands
from .registers import Registers
from .stdout import Stdout
from .variables import Variables




from __future__ import annotations
from typing import TYPE_CHECKING

from jpype import JImplements, JOverride

if TYPE_CHECKING:
    from java.awt.event import MouseEvent


@JImplements("java.awt.event.MouseListener")
class MouseListener:

    def __init__(self, clicked=None, pressed=None, released=None, entered=None, exited=None):
        self._clicked = clicked
        self._pressed = pressed
        self._released = released
        self._entered = entered
        self._exited = exited

    @JOverride
    def mouseClicked(self, event: MouseEvent):
        if self._clicked:
            self._clicked(event)

    @JOverride
    def mousePressed(self, event: MouseEvent):
        if self._pressed:
            self._pressed(event)

    @JOverride
    def mouseReleased(self, event: MouseEvent):
        if self._released:
            self._released(event)

    @JOverride
    def mouseEntered(self, event: MouseEvent):
        if self._entered:
            self._entered(event)

    @JOverride
    def mouseExited(self, event: MouseEvent):
        if self._exited:
            self._exited(event)



"""
Interface for creating a TableModel for java.
"""
from __future__ import annotations
from typing import List, TYPE_CHECKING

from jpype import JImplements, JOverride


if TYPE_CHECKING:
    from java.lang import Object
    from javax.swing.event import TableModelListener


@JImplements("javax.swing.table.TableModel")
class TableModel:

    def __init__(self, headers: List[str], data: List[List[str]]):
        self._headers = headers
        self._data = data

    @JOverride
    def getRowCount(self) -> int:
        return len(self._data)

    @JOverride
    def getColumnCount(self) -> int:
        return len(self._headers)

    @JOverride
    def getColumnName(self, col: int):
        return self._headers[col]

    @JOverride
    def getColumnClass(self, col: int):
        from java.lang import String
        return String

    @JOverride
    def isCellEditable(self, row: int, col: int) -> bool:
        return False

    @JOverride
    def getValueAt(self, row: int, col: int) -> Object:
        return self._data[row][col]

    @JOverride
    def setValueAt(self, value, row: int, col: int):
        self._data[row][col] = str(value)

    @JOverride
    def addTableModelListener(self, listener: TableModelListener):
        ...

    @JOverride
    def removeTableModelListener(self, listener: TableModelListener):
        ...



"""
Interface for instruction management.
"""
from __future__ import annotations

import logging
import warnings
from copy import deepcopy
from typing import List, TYPE_CHECKING

import dragodis.interface
from dragodis import NotExistError

from . import utils
from .operands import Operand
from .exceptions import EmulationError

if TYPE_CHECKING:
    from rugosa.emulation.cpu_context import ProcessorContext

logger = logging.getLogger(__name__)


class Instruction:
    """
    Wraps a dragodis Instruction object in order to represent the dynamic
    version of an instruction based on the connected processor context.
    """

    # Cache for keeping track of instructions and their operand indexes.
    _operand_indices = {}

    # Class used for creating new operands.
    _operand_class = Operand

    def __init__(self, cpu_context: ProcessorContext, ip):
        self.ip = ip
        self._cpu_context = cpu_context
        self.__insn = None

    def __deepcopy__(self, memo):
        # When we deep copy, clear out the __insn attribute so we don't
        # run into any serialization issues with bridged objects.
        deepcopy_method = self.__deepcopy__
        self.__deepcopy__ = None
        self.__insn = None
        copy = deepcopy(self, memo)
        self.__deepcopy__ = deepcopy_method
        return copy

    def __str__(self):
        return f"<{self.__class__.__name__} 0x{self.ip:08x} - {self.text}>"

    @property
    def _insn(self) -> dragodis.interface.Instruction:
        if not self.__insn:
            try:
                insn = self._cpu_context.emulator.disassembler.get_instruction(self.ip)
            except dragodis.NotExistError:
                # TODO: Should we just allow the NotExistError error to happen?
                raise EmulationError(f"Failed to get instruction at 0x{self.ip:X}")
            self.__insn = insn
        return self.__insn

    @property
    def data(self) -> bytes:
        """
        Bytes comprising the instruction
        """
        return self._insn.data

    @property
    def mnem(self) -> str:
        """Opcode mnemonic."""
        return self._insn.mnemonic

    # TODO: Move to dragodis
    @property
    def root_mnem(self) -> str:
        """
        Opcode mnemonic without any extensions such as condition codes, data type, etc.

        e.g.
            MOVSEQ -> MOV
        """
        return self._insn.root_mnemonic

    @property
    def text(self) -> str:
        """Disassembled code."""
        return self._insn.text

    @property
    def is_terminal(self) -> bool:
        return self._insn.is_return

    @property
    def is_call(self) -> bool:
        return self._insn.is_call

    @property
    def next_ip(self) -> int:
        """Obtains the address for the next instruction after this one."""
        return self._insn.line.next.address

    # TODO: Overwrite this in ARM to handle that special case.
    @property
    def operands(self) -> List[_operand_class]:
        return [
            self._operand_class(self._cpu_context, operand)
            for operand in self._insn.operands
        ]

    def _record_func_args(self):
        """
        Reset stack pointer reference and
        record function argument variables if we are executing the beginning of a function.
        """
        try:
            func_obj = self._cpu_context.emulator.disassembler.get_function(self.ip)
            if func_obj.start != self.ip:
                return
        except NotExistError:
            return

        # Reset the sp_start
        self._cpu_context._sp_start = self._cpu_context.sp

        # Add the passed in arguments to the variables map.
        for arg in self._cpu_context.passed_in_args:
            addr = arg.addr
            # TODO: Support variables from registers?
            if addr is not None:
                if arg.is_stack:
                    try:
                        stack_variable = func_obj.stack_frame[arg.name]
                        self._cpu_context.variables.add(addr, stack_variable)
                    except (KeyError, ValueError):
                        # TODO: passed in arguments aren't found on the stack for IDA.
                        #   This is due to IDA not adding argument names in function signature.
                        logger.warning(f"Failed to get stack information for function argument: {repr(arg)}")
                else:
                    self._cpu_context.variables.add(addr)

    def get_hooks(self, pre=True):
        """
        Retrieves callback hooks for the given instruction.
        :param pre: Whether to retrieve pre or post execution hooks.
        """
        return (
            self._cpu_context.emulator.get_instruction_hooks(self.ip, pre=pre)
            + self._cpu_context.emulator.get_instruction_hooks(self.mnem, pre=pre)
        )

    def _execute_hooks(self, pre=True):
        """
        Executes instructions hooks for the given start
        """
        for hook in self.get_hooks(pre=pre):
            try:
                hook(self._cpu_context, self)
            except RuntimeError:
                raise  # Allow RuntimeError exceptions to be thrown.
            except Exception as e:
                logger.debug("Failed to execute instruction hook with error: %s", e)

    def _execute(self):
        """
        Internal execute code used to execute the instruction itself and
        perform any custom tasks specific to the architecture.
        """
        emulator = self._cpu_context.emulator
        opcode_func = (
            emulator.get_opcode_hook(self.mnem)  # e.g. "bleq"
            or emulator.get_opcode_hook(self.root_mnem)  # e.g. "bl" from "bleq"
        )
        if opcode_func:
            opcode_func(self._cpu_context, self)
        else:
            logger.debug("%s instruction not implemented.", self.mnem)

    # TODO: All variable add_references() should be done here instead of in Operand class.
    def execute(self):
        """
        Emulate the instruction and store results in the context.
        """
        # Log a header line for debug messages of this instruction.
        # This is simpler and faster than trying to include the information at each log line
        logger.debug("[0x%X %03X] :: %s", self.ip, self._cpu_context.sp_diff, self.mnem)

        # Set instruction pointer to where we are currently executing.
        self._cpu_context.ip = self.ip

        # Extra processing if we are at the start of a function.
        self._record_func_args()

        # Run any pre-hooks first.
        self._execute_hooks(pre=True)

        # Execute the instruction.
        try:
            self._execute()
        except Exception:
            logger.exception("Failed to execute address 0x%X: %s", self.ip, self.text)

        # Record any variables encountered in the operands.
        for operand in self.operands:
            var = operand._operand.variable
            if var:
                self._cpu_context.variables.add(
                    operand.addr or operand.value, var, reference=self._cpu_context.ip
                )

        # Record executed instruction.
        self._cpu_context.executed_instructions.append(self.ip)

        # Run any post-hooks.
        self._execute_hooks(pre=False)

        # Add a blank space to help visually separate logs for each instruction.
        logger.debug("  ")

        # After execution, set instruction pointer to next instruction assuming
        # standard code flow and if no jump was made.
        if self._cpu_context.ip == self.ip:
            self._cpu_context.ip = self.next_ip

    def execute_call_hooks(self, func_name, func_ea):
        """
        Collect call history and emulates the affects of the function call
        by executing call hooks.

        :param func_name: Name of the function (or empty string)
        :param func_ea: Address of function to call.
        """
        warnings.warn("execute_call_hooks() has been moved to ProcessorContext._execute_call()", DeprecationWarning)
        self._cpu_context._execute_call(func_ea, func_name, self.ip)




from __future__ import annotations
from typing import TYPE_CHECKING, Dict, Optional

import rugosa

from dragodis.ghidra import GhidraLocal
from ghidra.util import Swing
from jpype import JImplements, JOverride
from rugosa.emulation.cpu_context import ProcessorContext


if TYPE_CHECKING:
    from ghidra.program.model.listing import Program
    from ghidra.program.util import ProgramLocation
    from ghidra.program.model.address import Address
    from rugosa.ghidra_plugin.components.emulator import EmulatorForm


class ProgramEmulationHelper:

    def __init__(self, program: Program):
        self._program = program
        self._dis = None
        self.emulator: rugosa.Emulator = None
        self.ctx: Optional[ProcessorContext] = None

    @property
    def dis(self) -> GhidraLocal:
        if self._dis is None:
            # lazy initialization
            self._dis = GhidraLocal(currentProgram=self._program)
            self.emulator = rugosa.Emulator(self._dis)
        return self._dis


@JImplements("dc3.rugosa.plugin.RugosaProgramListener")
class ProgramListener:

    def __init__(self, form: EmulatorForm):
        self.form = form
        self.currentProgram: Program = None
        self.location: ProgramLocation = None
        self.emulators: Dict[Program, ProgramEmulationHelper] = {}

    @JOverride
    def programOpened(self, program: Program):
        self.emulators[program] = ProgramEmulationHelper(program)

    @JOverride
    def programClosed(self, program: Program):
        helper = self.emulators.get(program, None)
        if helper is None:
            return
        helper.dis.stop()
        del self.emulators[program]

    @JOverride
    def programActivated(self, program: Program):
        self.currentProgram = program
        Swing.runIfSwingOrRunLater(self.form.update)

    @JOverride
    def programDeactivated(self, program: Program):
        self.currentProgram = None
        Swing.runIfSwingOrRunLater(self.form.update)

    @JOverride
    def locationChanged(self, loc: ProgramLocation):
        self.location = loc

    @property
    def currentAddress(self) -> Address:
        if self.location:
            return self.location.getAddress()
        return Address.NO_ADDRESS

    @property
    def state(self) -> ProgramEmulationHelper:
        if self.currentProgram is None:
            return None
        return self.emulators[self.currentProgram]



"""
Functions found in wd2_32.dll

The Winsock API
"""

import logging
import struct

from ...call_hooks import builtin_func
from ... import constants


logger = logging.getLogger(__name__)


@builtin_func("inet_addr")
#typespec("unsigned long inet_addr(const char *cp);")
def inet_addr(cpu_context, func_name, func_args):
    """
    Convert the provided inet_address to an IPv4 address
    """
    src = func_args[0]
    addr_str = cpu_context.memory.read_data(src, data_type=constants.STRING)
    if not addr_str:
        return 0
    try:
        addr = bytes(map(int, addr_str.decode().split('.')))
    except ValueError:
        logger.debug("Failed to convert value %s to an IPv4 address", addr_str)
        return 0
    return struct.unpack("<I", addr)[0]


@builtin_func("htons")
#typespec("u_short htons(u_short hostshort);")
def htons(cpu_context, func_name, func_args):
    """
    Convert the provided 16-bit number in host byte order (little-endian) to network byte order (big-endian)
    """
    le_port = func_args[0]
    port_data = struct.pack("<H", le_port)
    return struct.unpack(">H", port_data)[0]



"""
Interface for operand management in ARM.
"""

import logging
from typing import Optional, List

from dragodis.interface import Register, Phrase, MemoryReference, RegisterList
from dragodis.interface.types import ARMShiftType, OperandType

from ..operands import Operand
from ..exceptions import EmulationError
from .. import utils
from . import utils as arm_utils

logger = logging.getLogger(__name__)


class ARMOperand(Operand):

    _shift_map = {
        ARMShiftType.LSL: arm_utils.lsl,
        ARMShiftType.LSR: arm_utils.lsr,
        ARMShiftType.ASR: arm_utils.asr,
        ARMShiftType.ROR: arm_utils.ror,
        # rrx requires also passing in carry. (see usage)
        # TODO: Support other shift operations.
        ARMShiftType.UXTX: arm_utils.lsl
    }

    @property
    def offset(self) -> Optional[int]:
        """The offset value if the operand is a displacement/phrase."""
        phrase = self._operand.value
        if not isinstance(phrase, Phrase) or phrase.offset is None:
            return None

        offset = phrase.offset

        # Offset is a register.
        # [R1, R2]
        if isinstance(offset, Register):
            offset = self._cpu_context.registers[offset.name]
            # We could also have a shift applied in the offset.
            #   [R1, R2, LSL #3]
            offset = self._calc_shift(offset)
            offset = utils.signed(offset, self._cpu_context.bitness)

        # Otherwise offset is an immediate.
        # [R1, #1]

        return offset

    @property
    def shift_count(self) -> int:
        """The amount to shift (if a shifted register)"""
        _, shift_count = self._operand.shift
        if isinstance(shift_count, Register):
            shift_count = self._cpu_context.registers[shift_count.name]
        return shift_count

    def _calc_shift(self, value) -> int:
        """
        Calculates the shift applied within the operand.
        This could be applied directly or within the offset of a displacement.
            e.g.
                R2, LSL #3 -> R2 << 3
                R2, LSL R3 -> R2 << R3
                [R1, R2, LSL #3]  -> R2 << 3

        NOTE: Any modifications to the carry flag will be applied only if
             the condition flag is set and the context's ip is the same as the
             address of the operand's instruction.

        :param value: The base value the shift is to be applied to.
        :return: Results of the shift (or original value back if no shift is applicable.)
        """
        count = self.shift_count
        if count > 0:
            shift_op, _ = self._operand.shift
            if shift_op == ARMShiftType.RRX:  # RRX also requires original carry flag
                carry, value = arm_utils.rrx(self._cpu_context.registers.c, value, count)
            else:
                carry, value = self._shift_map[shift_op](value, count)

            # Update carry flag if condition flag is set for instruction (S postfix)
            # (But only update if context's instruction pointer is still looking at this instruction.)
            if self._operand.instruction.update_flags and self._cpu_context.ip == self.ip:
                self._cpu_context.registers.c = carry

        return value

    @property
    def addr(self) -> Optional[int]:
        """
        The referenced memory address of the operand.
        :return int: Memory address or None if operand is not a memory reference.
        """
        addr = None
        value = self._operand.value

        if isinstance(value, Phrase):
            addr = self.base

            # Ignore including the offset if post indexed.
            #   ie. include offset for [R2, #4] but ignore for [R2], #4
            if not self._operand.instruction.post_indexed:
                offset = self.offset
                logger.debug("0x%X + 0x%X = 0x%X", addr, offset, addr + offset)
                addr += offset

            if addr < 0:
                logger.debug("Address is negative, resorting to address of 0.")
                addr = 0

        elif isinstance(value, MemoryReference):
            addr = int(value)

        if addr is not None:
            logger.debug("&%s -> 0x%X", self.text, addr)

        return addr

    @property
    def is_signed(self) -> bool:
        """
        Whether the memory addressing is signed.
        """
        mnem = self._operand.instruction.mnemonic
        return any(suffix in mnem for suffix in ["sb", "sh", "sw"])

    @property
    def register_list(self) -> Optional[List[str]]:
        """
        List of register names if operand is a register list.
        """
        reg_list = self._operand.value
        if not isinstance(reg_list, RegisterList):
            return None
        return [reg.name for reg in reg_list]

    @property
    def value(self):
        value = self._operand.value

        # Barrel shifter
        if isinstance(value, Register):
            value = self._cpu_context.registers[value.name]
            # Run _calc_shift to handle barrel shifter.
            # (Shift won't occur if shift count is 0)
            value = self._calc_shift(value)
            self._record_address_variable(value)
            return value

        # Register list
        if isinstance(value, RegisterList):
            return [self._cpu_context.registers[reg.name] for reg in value]

        value = super().value

        # If a memory reference, the final value may be signed.
        if isinstance(value, (MemoryReference, Phrase)) and self.is_signed:
            value = utils.signed(value, self._cpu_context.byteness)

        return value

    @value.setter
    def value(self, value):
        try:
            logger.debug("0x%X -> %s", value, self.text)
        except TypeError:
            logger.debug("%r -> %s", value, self.text)

        # Barrel shifter
        if self.type == OperandType.register and self.shift_count:
            raise EmulationError(f"Unable to set value to operand with a shift", ip=self.ip)

        # Register list
        reg_list = self.register_list
        if reg_list:
            # To set the value for an operand that is a register list, user must provide
            # a list of value of equal size.
            # User can use None to indicate not to update that register.
            if not isinstance(value, list) or len(value) != len(reg_list):
                raise ValueError(f"Operand value for {self.text} must be a list of {len(reg_list)} values.")
            for reg_name, reg_value in zip(reg_list, value):
                if reg_value is not None:
                    self._cpu_context.registers[reg_name] = reg_value
            return

        super(ARMOperand, self.__class__).value.__set__(self, value)




from pathlib import Path

import pyhidra

from rugosa import __version__
from jpype.types import JClass


def setup(launcher):
    """
    Run by pyhidra launcher to install our plugin.
    """
    source_path = Path(__file__).parent / "java" / "plugin"
    details = pyhidra.ExtensionDetails(
        name="rugosa",
        description="Rugosa Emulator Plugin",
        author="Department of Defence Cyber Crime Center (DC3)",
        plugin_version=__version__,
    )
    launcher.install_plugin(source_path, details)


def pre_launch():
    # rugosa classes might not exist when this module is imported
    # the emulator functions must be imported here
    from java.lang import ClassLoader
    from rugosa.ghidra_plugin.components.emulator import initialize

    gcl = ClassLoader.getSystemClassLoader()
    RugosaPlugin = JClass("dc3.rugosa.plugin.RugosaPlugin", loader=gcl)
    RugosaPlugin.setInitiailizer(initialize)



from __future__ import annotations
from typing import TYPE_CHECKING

import ida_kernwin
from PyQt5 import QtWidgets

if TYPE_CHECKING:
    from rugosa.ida_plugin.components.emulator import EmulatorForm


class Stdout:

    def __init__(self, ui: EmulatorForm):
        self._ctx = None
        self.ui = ui
        self.textdump = ui.stdout_textdump
        self.export_button = ui.stdout_export_button
        self.export_button.clicked.connect(self.export)

    def export(self):
        """
        Exports current stdout dump to a text file.
        """
        if not self._ctx:
            return
        file_path = ida_kernwin.ask_file(True, "*.txt", "Dump Stdout")
        if file_path:
            with open(file_path, "w") as outfile:
                outfile.write(self._ctx.stdout)

    def populate(self, ctx):
        """
        Populate the stdout text box with current stdout in context.
        """
        self._ctx = ctx
        self.textdump.setText(f"<pre>{ctx.stdout}</pre>")



from __future__ import annotations
from typing import TYPE_CHECKING

import jpype

RugosaActionListener = jpype.JClass("dc3.rugosa.plugin.RugosaActionListener")

from rugosa.ghidra_plugin.gui.error_msg import catch_errors
from rugosa.ghidra_plugin.gui.table import TableModel
from rugosa.ghidra_plugin.gui.mouse_listener import MouseListener


if TYPE_CHECKING:
    from java.awt.event import MouseEvent
    from rugosa.ghidra_plugin.components.emulator import EmulatorForm


class Actions:

    SIDE_HEADERS = ["Order", "Address", "Action"]
    FIELD_HEADERS = ["Name", "Value"]

    def __init__(self, form: EmulatorForm):
        self._ctx = None
        self._actions = None
        self.form = form
        self.ui = form.ui
        self.fields_table = self.ui.actionsFieldsTable
        self.side_table = self.ui.actionsSideTable
        # TODO: Figure out how to make a selection listener instead.
        self.side_table.addMouseListener(MouseListener(clicked=self.side_table_clicked))

    @catch_errors
    def side_table_clicked(self, event: MouseEvent):
        if not self._ctx:
            return
        action = self._actions[self.side_table.getSelectedRow()]
        self.populate_fields_table(action)

    @catch_errors
    def populate_fields_table(self, action):
        rows = []
        for name, value in action:
            if name != "ip":
                rows.append([name, str(value)])
        self.fields_table.setModel(TableModel(self.FIELD_HEADERS, rows))

    @catch_errors
    def populate(self, ctx):
        """
        Populates the variables table.
        """
        self._ctx = ctx
        self._actions = list(reversed(ctx.actions))

        rows = []
        for index, action in enumerate(self._actions):
            rows.append([str(index), hex(action.ip), action.__class__.__name__])
        self.side_table.setModel(TableModel(self.SIDE_HEADERS, rows))




from __future__ import annotations
from typing import TYPE_CHECKING

import ida_kernwin
import ida_bytes
from PyQt5 import QtWidgets

if TYPE_CHECKING:
    from rugosa.ida_plugin.components.emulator import EmulatorForm


class Operands:

    def __init__(self, ui: EmulatorForm):
        self._ctx = None
        self.ui = ui
        self.table = ui.operands_table
        self.table.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        self.table.itemDoubleClicked.connect(ui.redirect)
        self.table.customContextMenuRequested.connect(self.context_menu)

    def context_menu(self, pos):
        if not self._ctx:
            return
        index = self.table.indexAt(pos)
        if not index.isValid():
            return
        operand = self._ctx.operands[self.table.rowAt(pos.y())]
        address = operand.addr or operand.value

        menu = QtWidgets.QMenu(self.table)
        jump = menu.addAction("Jump to disassembly")
        jump.setEnabled(ida_bytes.is_loaded(address))
        show_in_memory = menu.addAction("Show in Memory")
        show_in_memory.setEnabled(self._ctx.memory.is_mapped(address))

        action = menu.exec_(self.table.viewport().mapToGlobal(pos))
        if action == jump:
            ida_kernwin.jumpto(address)
        elif action == show_in_memory:
            # Size of pointed to value is unknown, so just showing the first 256 bytes.
            self.ui.memory.show_with_data(address, 256)

    def populate(self, ctx):
        """
        Populates the operands table.
        """
        self._ctx = ctx
        table = self.table
        operands = ctx.operands
        table.setRowCount(len(operands))
        for index, operand in enumerate(operands):
            table.setItem(index, 0, QtWidgets.QTableWidgetItem(str(operand.idx)))
            table.setItem(index, 1, QtWidgets.QTableWidgetItem(str(operand.width)))
            table.setItem(index, 2, QtWidgets.QTableWidgetItem(operand.text))
            if operand.addr is not None:
                table.setItem(index, 3, QtWidgets.QTableWidgetItem(hex(operand.addr)))
            else:
                table.setItem(index, 3, QtWidgets.QTableWidgetItem())
            value = operand.value
            if isinstance(value, int):
                value = hex(value)
            table.setItem(index, 4, QtWidgets.QTableWidgetItem(str(value)))

            # If value is a pointer, show the first few bytes of the referenced data
            value = operand.value
            if self._ctx.memory.is_mapped(value):
                data = self._ctx.memory.read(value, 24)
                table.setItem(index, 5, QtWidgets.QTableWidgetItem(str(data) + "..."))
            else:
                table.setItem(index, 5, QtWidgets.QTableWidgetItem())



"""
Tests utility functions.
"""

from rugosa.emulation import utils


def test_signed():
    assert utils.signed(1, 4) == 1
    assert utils.signed(0xffffff10, 4) == -240


def test_unsigned():
    assert utils.unsigned(1, 4) == 1
    assert utils.unsigned(-240, 4) == 0xffffff10


def test_sign_bit():
    assert utils.sign_bit(1, 4) == 0
    assert utils.sign_bit(0xffffff10, 4) == 1


def test_sign_extend():
    assert utils.sign_extend(1, 4, 8) == 1
    assert utils.sign_extend(0xffffff10, 4, 8) == 0xffffffffffffff10


def test_float_to_int():
    assert utils.float_to_int(1.0) == 0x3ff0000000000000


def test_int_to_float():
    assert utils.int_to_float(0x3ff0000000000000) == 1.0


def test_get_mask():
    assert utils.get_mask(4) == 0xffffffff


def test_sanitize_func_name():
    assert utils.sanitize_func_name("sprintf") == "sprintf"
    assert utils.sanitize_func_name("_sprintf") == "sprintf"
    assert utils.sanitize_func_name("_sprintf_2") == "sprintf"


def test_is_func_ptr(disassembler):
    assert utils.is_func_ptr(disassembler, 0x401000)
    assert utils.is_func_ptr(disassembler, 0x40A040)
    assert not utils.is_func_ptr(disassembler, 0x1)




import os
import pathlib
import sys
from typing import Optional


def get_ida_user_dir() -> Optional[pathlib.Path]:
    """
    Attempts to get the user directory for IDA.
    """
    # First check if IDAUSR has been set.
    path = os.getenv("IDAUSR")
    if path:
        return pathlib.Path(path)

    # Otherwise obtain path from preset based on IDA documentation.
    if sys.platform == "win32":
        appdata = os.getenv("APPDATA")
        if not appdata:
            return None
        return pathlib.Path(appdata, "Hex-Rays", "IDA Pro")
    else:
        return pathlib.Path("~", ".idapro").expanduser()



from __future__ import annotations
from typing import TYPE_CHECKING

import ida_kernwin
import ida_bytes
from PyQt5 import QtWidgets

if TYPE_CHECKING:
    from rugosa.ida_plugin.components.emulator import EmulatorForm


class Variables:

    MAX_VALUE_SIZE = 256

    def __init__(self, ui: EmulatorForm):
        self._ctx = None
        self._variables = None
        self.ui = ui
        self.table = ui.variables_table
        self.table.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        self.table.itemDoubleClicked.connect(ui.redirect)
        self.table.customContextMenuRequested.connect(self.context_menu)

    def context_menu(self, pos):
        if not self._variables:
            return
        index = self.table.indexAt(pos)
        if not index.isValid():
            return
        var = self._variables[self.table.rowAt(pos.y())]

        menu = QtWidgets.QMenu(self.table)
        jump = menu.addAction("Jump to disassembly")
        jump.setEnabled(ida_bytes.is_loaded(var.addr))
        show_in_memory = menu.addAction("Show in Memory")
        show_in_memory.setEnabled(self._ctx.memory.is_mapped(var.addr))

        action = menu.exec_(self.table.viewport().mapToGlobal(pos))
        if action == jump:
            ida_kernwin.jumpto(var.addr)
        elif action == show_in_memory:
            self.ui.memory.show_with_data(var.addr, var.size)

    def populate(self, ctx):
        """
        Populate variables table
        """
        self._ctx = ctx
        table = self.table
        # Ignore " r" variable. That is a proprietary IDA thing just to keep track of return variables.
        self._variables = [var for var in ctx.variables if var.name != " r"]
        table.setRowCount(len(self._variables))
        for index, var in enumerate(self._variables):
            table.setItem(index, 0, QtWidgets.QTableWidgetItem(hex(var.addr)))
            table.setItem(index, 1, QtWidgets.QTableWidgetItem("" if var.stack_offset is None else hex(var.stack_offset)))
            data_type = var.data_type
            if var.count > 1 and data_type != "func_ptr":
                data_type += f"[{var.count}]"
            table.setItem(index, 2, QtWidgets.QTableWidgetItem(data_type))
            table.setItem(index, 3, QtWidgets.QTableWidgetItem(str(var.size)))
            table.setItem(index, 4, QtWidgets.QTableWidgetItem(var.name))
            value = var.value
            if isinstance(value, (list, bytes)) and len(value) > self.MAX_VALUE_SIZE:
                value = str(value[:self.MAX_VALUE_SIZE]) + " (truncated)"
            else:
                value = str(value)
            table.setItem(index, 5, QtWidgets.QTableWidgetItem(value))


from __future__ import annotations
from typing import TYPE_CHECKING

import jpype

RugosaActionListener = jpype.JClass("dc3.rugosa.plugin.RugosaActionListener")

from rugosa.ghidra_plugin.gui.error_msg import catch_errors
from rugosa.ghidra_plugin.gui.table import TableModel
from rugosa.ghidra_plugin.gui.context_menu import TableContextMenuEntry


if TYPE_CHECKING:
    from rugosa.ghidra_plugin.components.emulator import EmulatorForm


class ContextMenuEntry(TableContextMenuEntry):

    def __init__(self, operands: Operands):
        super().__init__(operands.table)
        self._operands = operands

    @property
    def address(self):
        operand = self._operands._ctx.operands[self._component.getSelectedRow()]
        return operand.addr or operand.value


class JumpToDisassembly(ContextMenuEntry):
    NAME = "Jump To Disassembly"

    def enabled(self, row: int, col: int) -> bool:
        ctx = self._operands._ctx
        address = self.address
        return bool(ctx and address)

    @catch_errors
    def clicked(self, row: int, col: int):
        self._operands.form.plugin.goTo(self._operands.form.dis._to_addr(self.address))


class ShowInMemory(ContextMenuEntry):
    NAME = "Show in Memory"

    @catch_errors
    def enabled(self, row: int, col: int) -> bool:
        ctx = self._operands._ctx
        address = self.address
        return bool(ctx and address and ctx.memory.is_mapped(address))

    @catch_errors
    def clicked(self, row: int, col: int):
        self._operands.form.memory.show_with_data(self.address, 256)


class Operands:

    HEADERS = ["Index", "Size", "Text", "Address", "Value", "Referenced Data"]

    def __init__(self, form: EmulatorForm):
        self._ctx = None
        self.form = form
        self.ui = form.ui
        self.table = self.ui.operandsTable
        self.form.add_context_menu_entry(JumpToDisassembly(self))
        self.form.add_context_menu_entry(ShowInMemory(self))

    @catch_errors
    def populate(self, ctx):
        """
        Populates the operands table.
        """
        self._ctx = ctx
        operands = ctx.operands

        rows = []
        for operand in operands:
            row = [""] * len(self.HEADERS)
            row[0] = str(operand.idx)
            row[1] = str(operand.width)
            row[2] = operand.text

            if operand.addr is not None:
                row[3] = hex(operand.addr)

            value = operand.value
            if isinstance(value, int):
                value = hex(value)
            row[4] = str(value)

            # If value is a pointer, show the first few bytes of the referenced data
            value = operand.value
            if self._ctx.memory.is_mapped(value):
                data = self._ctx.memory.read(value, 24)
                row[5] = str(data) + "..."

            rows.append(row)

        # TODO: Add table listeners
        self.table.setModel(TableModel(self.HEADERS, rows))



"""
Rugosa IDA Plugin.
"""

import ida_idaapi

import dragodis

import rugosa
from rugosa.ida_plugin.components.emulator import EmulatorForm


class RugosaPlugin(ida_idaapi.plugin_t):
    flags = ida_idaapi.PLUGIN_FIX
    comment = "Rugosa"
    help = "See https://github.com/dod-cyber-crime-center/rugosa/blob/master/readme.md"
    wanted_name = "Rugosa Emulator"
    wanted_hotkey = "Ctrl-Alt-R"

    def init(self):
        """Called when plugin is initially loaded."""
        print(
            f"-------------------------------\n"
            f"Rugosa Emulator\n"
            f"rugosa v{rugosa.__version__}, dragodis v{dragodis.__version__}\n"
            f"-------------------------------\n"
        )
        self.form = None
        return ida_idaapi.PLUGIN_KEEP

    def run(self, arg=0):
        """Starts the plugin."""
        if not self.form:
            self.form = EmulatorForm(dragodis.IDA())
        self.form.Show(self.wanted_name)

    def term(self):
        ...



"""
Functions found in shell32.dll
"""

import logging

from . import win_constants as wc
from ...call_hooks import builtin_func
from ... import actions

logger = logging.getLogger(__name__)


@builtin_func("ShellExecuteA")
@builtin_func("ShellExecuteW")
#typespec("HINSTANCE ShellExecuteA(HWND hwnd, LPCSTR lpOperation, LPCSTR lpFile, LPCSTR lpParameters, LPCSTR lpDirectory, INT nShowCmd);)")
def shell_execute(cpu_context, func_name, func_args):
    """
    Performs an operation on a specified file
    """
    wide = func_name.endswith("W")

    _, operation_ptr, file_ptr, params_ptr, dir_ptr, visibility = func_args

    filepath = cpu_context.memory.read_string(file_ptr, wide=wide)
    params = ""
    operation = ""
    directory = ""

    if params_ptr:
        params = cpu_context.memory.read_string(params_ptr, wide=wide)

    if operation_ptr:
        operation = cpu_context.memory.read_string(operation_ptr, wide=wide)

    if dir_ptr:
        directory = cpu_context.memory.read_string(dir_ptr, wide=wide)

    logger.debug("%s: %r %r", func_name, operation, filepath)
    cpu_context.actions.add(
        actions.ShellOperation(cpu_context.ip, operation, filepath, params, directory, wc.Visibility(visibility))
    )

    return cpu_context.objects.alloc()


# Dictionary mapping CSIDL values to their default values on 64-bit Windows 10. Created by taking the listing from the
# SpecialFoldersView.exe utility from NirSoft and then cleaning it up to use more appropriate environment variables.
CSIDL_DICT_WIN10x64 = {
    wc.CSIDL.CSIDL_PROGRAMS: r"%AppData%\Microsoft\Windows\Start Menu\Programs",
    wc.CSIDL.CSIDL_PERSONAL: r"%UserProfile%\Documents",
    wc.CSIDL.CSIDL_FAVORITES: r"%UserProfile%\Favorites",
    wc.CSIDL.CSIDL_STARTUP: r"%AppData%\Microsoft\Windows\Start Menu\Programs\Startup",
    wc.CSIDL.CSIDL_RECENT: r"%AppData%\Microsoft\Windows\Recent",
    wc.CSIDL.CSIDL_SENDTO: r"%AppData%\Microsoft\Windows\SendTo",
    wc.CSIDL.CSIDL_STARTMENU: r"%AppData%\Microsoft\Windows\Start Menu",
    wc.CSIDL.CSIDL_MYMUSIC: r"%UserProfile%\Music",
    wc.CSIDL.CSIDL_MYVIDEO: r"%UserProfile%\Videos",
    wc.CSIDL.CSIDL_DESKTOPDIRECTORY: r"%UserProfile%\Desktop",
    wc.CSIDL.CSIDL_NETHOOD: r"%AppData%\Microsoft\Windows\Network Shortcuts",
    wc.CSIDL.CSIDL_FONTS: r"%SystemRoot%\Fonts",
    wc.CSIDL.CSIDL_TEMPLATES: r"%AppData%\Microsoft\Windows\Templates",
    wc.CSIDL.CSIDL_COMMON_STARTMENU: r"%AllUsersProfile%\Microsoft\Windows\Start Menu",
    wc.CSIDL.CSIDL_COMMON_PROGRAMS: r"%AllUsersProfile%\Microsoft\Windows\Start Menu\Programs",
    wc.CSIDL.CSIDL_COMMON_STARTUP: r"%AllUsersProfile%\Microsoft\Windows\Start Menu\Programs\Startup",
    wc.CSIDL.CSIDL_COMMON_DESKTOPDIRECTORY: r"%Public%\Desktop",
    wc.CSIDL.CSIDL_APPDATA: r"%AppData%",
    wc.CSIDL.CSIDL_PRINTHOOD: r"%AppData%\Microsoft\Windows\Printer Shortcuts",
    wc.CSIDL.CSIDL_LOCAL_APPDATA: r"%LocalAppData%",
    wc.CSIDL.CSIDL_COMMON_FAVORITES: r"%UserProfile%\Favorites",
    wc.CSIDL.CSIDL_INTERNET_CACHE: r"%LocalAppData%\Microsoft\Windows\INetCache",
    wc.CSIDL.CSIDL_COOKIES: r"%LocalAppData%\Microsoft\Windows\INetCookies",
    wc.CSIDL.CSIDL_HISTORY: r"%LocalAppData%\Microsoft\Windows\History",
    wc.CSIDL.CSIDL_COMMON_APPDATA: r"%AllUsersProfile%",
    wc.CSIDL.CSIDL_WINDOWS: r"%WinDir%",
    wc.CSIDL.CSIDL_SYSTEM: r"%SystemRoot%\System32",
    wc.CSIDL.CSIDL_PROGRAM_FILES: r"%ProgramFiles%",
    wc.CSIDL.CSIDL_MYPICTURES: r"%UserProfile%\Pictures",
    wc.CSIDL.CSIDL_PROFILE: r"%UserProfile%",
    wc.CSIDL.CSIDL_SYSTEMX86: r"%SystemRoot%\SysWOW64",
    wc.CSIDL.CSIDL_PROGRAM_FILESX86: r"%ProgramFiles(x86)%",
    wc.CSIDL.CSIDL_PROGRAM_FILES_COMMON: r"%CommonProgramFiles%",
    wc.CSIDL.CSIDL_PROGRAM_FILES_COMMONX86: r"%CommonProgramFiles(x86)%",
    wc.CSIDL.CSIDL_COMMON_TEMPLATES: r"%AllUsersProfile%\Microsoft\Windows\Templates",
    wc.CSIDL.CSIDL_COMMON_DOCUMENTS: r"%Public%\Documents",
    wc.CSIDL.CSIDL_COMMON_ADMINTOOLS: r"%AllUsersProfile%\Microsoft\Windows\Start Menu\Programs\Administrative Tools",
    wc.CSIDL.CSIDL_ADMINTOOLS: r"%AppData%\Microsoft\Windows\Start Menu\Programs\Administrative Tools",
    wc.CSIDL.CSIDL_COMMON_MUSIC: r"%Public%\Music",
    wc.CSIDL.CSIDL_COMMON_PICTURES: r"%Public%\Pictures",
    wc.CSIDL.CSIDL_COMMON_VIDEO: r"%Public%\Videos",
    wc.CSIDL.CSIDL_CDBURN_AREA: r"%LocalAppData%\Microsoft\Windows\Burn\Burn"
}


@builtin_func("SHGetFolderPathA")
@builtin_func("SHGetFolderPathW")
@builtin_func("SHGetSpecialFolderPathA")
@builtin_func("SHGetSpecialFolderPathW")
#typespec("SHFOLDERAPI SHGetFolderPathA(HWND hwnd, int csidl, HANDLE hToken, DWORD dwFlags,  LPSTR  pszPath);")
#typespec("BOOL SHGetSpecialFolderPathA(HWND hwnd, LPSTR pszPath, int csidl, BOOL fCreate);")
def sh_get_folder_path(cpu_context, func_name, func_args):
    """
    Retrieves the CSIDL folder mapping for the provided CSIDL Enumeration

    Attempts to expand folder based on default 64-bit Windows 10 values.
    """
    wide = func_name.endswith("W")
    if "Special" in func_name:
        _, path_ptr, csidl, _ = func_args
    else:
        _, csidl, _, _, path_ptr = func_args

    # Mask off any CSIDL flags
    csidl &= 0xff
    if csidl in CSIDL_DICT_WIN10x64:
        folder = CSIDL_DICT_WIN10x64[csidl]
    else:
        logger.warning("Could not acquire a folder for CSIDL %d", csidl)
        folder = f"{csidl}"

    logger.debug("Writing CSIDL %s to 0x%08X", folder, path_ptr)
    cpu_context.memory.write_string(path_ptr, folder, wide=wide)
    return wc.ERROR_SUCCESS



"""
Interface for operand management.
"""

from __future__ import annotations
import collections
import logging
from copy import deepcopy
from typing import TYPE_CHECKING, Optional, Union

# TODO: Should dragodis have it's interface in the root module?
import dragodis.interface
from dragodis import OperandType
from dragodis.interface import Phrase, Immediate, Register, MemoryReference

from . import utils
from .exceptions import EmulationError

if TYPE_CHECKING:
    from .cpu_context import ProcessorContext


logger = logging.getLogger(__name__)


class Operand:
    """
    Stores information for a given operand for a specific CPU context state.
    Wraps a dragodis Operand object in order to provide dynamic information
    for the operand based on the attached processor context.
    """

    def __init__(self, cpu_context: ProcessorContext, operand: dragodis.interface.Operand):
        """
        :param cpu_context: CPU context to pull operand value
        :param operand: dragodis Operand object to wrap.
        """
        self._cpu_context = cpu_context
        self._operand = operand

    def __repr__(self):
        string = f"<{self.__class__.__name__} 0x{self.ip:0x}:{self.idx} : {self.text} = {self.value!r}"
        if self.addr is not None:
            string += f" : &{self.text} = 0x{self.addr:0x}"
        string += f" : width = {self.width}>"
        return string

    def __deepcopy__(self, memo):
        # When we deep copy, clear out the __insn attribute so we don't
        # run into any serialization issues with Swig objects.
        deepcopy_method = self.__deepcopy__
        self.__deepcopy__ = None
        self.__insn = None
        copy = deepcopy(self, memo)
        self.__deepcopy__ = deepcopy_method
        return copy

    @property
    def ip(self) -> int:
        return self._operand.address

    @property
    def idx(self) -> int:
        return self._operand.index

    # TODO: Update those that use this attribute.
    @property
    def type(self) -> dragodis.interface.OperandType:
        return self._operand.type

    @property
    def text(self) -> str:
        return self._operand.text

    def _record_address_variable(self, addr: int):
        """
        Record the reference to the address if address is a defined variable.
        """
        if addr in self._cpu_context.variables:
            self._cpu_context.variables[addr].add_reference(self.ip)

    @property
    def width(self):
        """
        Based on the dtype value, the size of the operand in bytes

        :return: size of data type
        """
        return self._operand.width

    @property
    def is_hidden(self):
        """
        True if the operand is not part of the visible assembly code.
        (These are for implicit registers like EAX)
        """
        return self._operand.text == "" or self._operand.type == OperandType.void

    @property
    def is_func_ptr(self):
        """True if the operand is a pointer to a function."""
        return utils.is_func_ptr(self._cpu_context.emulator.disassembler, self.addr or self.value)

    @property
    def offset(self) -> Optional[int]:
        """The offset value if the operand is a displacement."""
        return None

    @property
    def base(self) -> Optional[int]:
        """The value of the base register if operand is a displacement."""
        phrase = self._operand.value
        if not isinstance(phrase, Phrase):
            return None
        base_reg = phrase.base
        if not base_reg:
            return 0
        value = self._cpu_context.registers[base_reg.name]
        return utils.signed(value, self._cpu_context.byteness)

    @base.setter
    def base(self, value: int):
        """Sets the value of the base register if operand is a displacement."""
        phrase = self._operand.value
        if isinstance(phrase, Phrase):
            self._cpu_context.registers[phrase.base.name] = value

    @property
    def addr(self) -> Optional[int]:
        """
        Retrieves the referenced memory address of the operand.

        This should be overwritten by architecture specific Operand implementations
        if this property is applicable.

        :return int: Memory address or None if operand is not a memory reference.
        """
        return None

    @property
    def value(self) -> Union[None, int, bytes]:
        """
        Retrieve the value of the operand as it is currently in the cpu_context.
        NOTE: We can't cache this value because the value may change based on the cpu context.

        :return int: An integer or byes of the operand value.
        """
        if self.is_hidden:
            return None

        value = self._operand.value

        if isinstance(value, Immediate):
            return int(value)

        # Record reference if address is a variable address and memory reference.
        # operand is one of o_near, o_far, o_mem
        if isinstance(value, MemoryReference) and self._operand.type == OperandType.code:
            addr = int(value)
            self._record_address_variable(addr)
            return addr

        if isinstance(value, Register):
            value = self._cpu_context.registers[value.name]
            self._record_address_variable(value)
            return value

        if isinstance(value, (Phrase, MemoryReference)):
            addr = self.addr
            self._record_address_variable(addr)

            # If a function pointer, we want to return the address.
            # This is because a function may be seen as a memory reference, but we don't
            # want to dereference it in case it in a non-call instruction.
            # (e.g.  "mov  esi, ds:LoadLibraryA")
            # NOTE: Must use internal function to avoid recursive loop.
            if utils.is_func_ptr(self._cpu_context.emulator.disassembler, addr):
                return addr

            # Return empty
            if not self.width:
                logger.debug("Width is zero for %s, returning empty string.", self.text)
                return b""

            # Otherwise, dereference the address.
            value = self._cpu_context.memory.read(addr, self.width)
            return int.from_bytes(value, self._cpu_context.byteorder)

        raise EmulationError(f"Invalid operand type: {self.type}", ip=self.ip)

    @value.setter
    def value(self, value: Union[int, bytes]):
        """
        Set the operand to the specified value within the cpu_context.
        """
        # Value may be signed.
        if isinstance(value, int) and value < 0:
            value = utils.unsigned(value, self.width)

        # If we are writing to an immediate, I believe they want to write to the memory at the immediate.
        # TODO: Should we fail instead?
        if self.type == OperandType.immediate:
            offset = self.value
            line = self._cpu_context.emulator.disassembler.get_line(offset)
            if line.is_loaded:
                self._cpu_context.memory.write(offset, value)
            return

        if self.type == OperandType.register:
            # Convert the value from string to integer...
            if isinstance(value, bytes):
                value = int.from_bytes(value, self._cpu_context.byteorder)
            reg = self._operand.value
            self._cpu_context.registers[reg.name] = value
            return

        if self.type in (OperandType.memory, OperandType.phrase):
            if isinstance(value, int):
                value = value.to_bytes(self.width, self._cpu_context.byteorder)
            self._cpu_context.memory.write(self.addr, value)
            return

        raise EmulationError(f"Invalid operand type: {self.type}", ip=self.ip)


# This is a "lite" version of the Operand class that only allows access only to a few attributes, is read only,
# and not backed by a CPU context.
OperandLite = collections.namedtuple("OperandLite", "ip idx text value")



"""
Import built-in function code based on input file.

These functions are used to emulate the effects of known builtin functions.

Add any builtin functions that need to be handled below.  The function should be declared as such

# Using the same function for multiple instructions:
@builtin_func("memmove")
@builtin_func("memcpy")
def _memcpy(cpu_context, call_ip, func_name, func_args):
    print "IN memmove or memcpy"
    return 1  # Return anything to be placed into rax (or equivalent)

# Using a single function for a builtin
@builtin_func
def memmove(cpu_context, call_ip, func_name, func_args):
    print "IN memmove"

"""

from ..registry import registrar

# Dictionary containing builtin function names -> function
BUILTINS = {}

# Dictionary containing builtin function names -> function
builtin_func = registrar(BUILTINS, name="builtin")


from . import stdlib

# TODO: We shouldn't be determining what global callhooks are available by import.
#   Instead generate a function that will produce the correct callhooks based
#   on the libraries detected in the disassembler.
#   Import based code flow is a no-no

# For now, import WinAPI always.
from . import win_api

# # TODO: Call the framework file type information rather than the IDA functions.
# import idc
#
# file_type = idc.get_inf_attr(idc.INF_FILETYPE)
# if file_type == idc.FT_PE:
#     from . import win_api
# elif file_type == idc.FT_ELF:
#     pass



import functools
import logging
import traceback

import jpype

logger = logging.getLogger(__name__)


def show_exception(exception: Exception, title="Error", level=logging.ERROR):
    """
    Shows exception in dialog box.
    """
    # Show error in console.
    logger.log(level, exception, exc_info=level > logging.WARNING)
    # Also show a popup dialog.
    details = str(exception)
    plugin = jpype.JClass("dc3.rugosa.plugin.RugosaPlugin")
    from ghidra.util import Msg
    if level <= logging.INFO:
        Msg.showInfo(plugin, None, title, details)
    elif level <= logging.WARNING:
        Msg.showWarn(plugin, None, title, details)
    else:
        # Show full traceback on error.
        details = "".join(traceback.format_tb(exception.__traceback__)) + f"\n{details}"
        Msg.showError(plugin, None, title, details)


def catch_errors(func):
    """
    Wraps a function to ensure any uncaught exceptions get shown as an error message
    in the Ghidra GUI instead of being thrown.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            show_exception(e)
    return wrapper



from __future__ import annotations
from typing import TYPE_CHECKING

import ida_kernwin
import ida_bytes
import dragodis
from PyQt5 import QtWidgets

if TYPE_CHECKING:
    from rugosa.ida_plugin.components.emulator import EmulatorForm


class FunctionArguments:

    def __init__(self, ui: EmulatorForm):
        self._ctx = None
        self._arguments = None
        self.ui = ui
        self.table = ui.function_arguments_table
        self.table.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        self.table.itemDoubleClicked.connect(ui.redirect)
        self.table.customContextMenuRequested.connect(self.context_menu)
        self.signature = ui.function_signature
        self.num_args = ui.num_args
        self.num_args.valueChanged.connect(self.num_args_updated)

    def context_menu(self, pos):
        if not self._arguments:
            return
        index = self.table.indexAt(pos)
        if not index.isValid():
            return
        arg = self._arguments[self.table.rowAt(pos.y())]
        value = arg.value

        menu = QtWidgets.QMenu(self.table)
        jump = menu.addAction("Jump to disassembly")
        jump.setEnabled(ida_bytes.is_loaded(value))
        show_in_memory = menu.addAction("Show in Memory")
        show_in_memory.setEnabled(self._ctx.memory.is_mapped(value))

        action = menu.exec_(self.table.viewport().mapToGlobal(pos))
        if action == jump:
            ida_kernwin.jumpto(value)
        elif action == show_in_memory:
            # Size of pointed to value is unknown, so just showing the first 256 bytes.
            self.ui.memory.show_with_data(value, 256)

    def num_args_updated(self, value: int):
        if self._ctx:
            self.populate(self._ctx, num_args=value)

    def _populate_table(self, arguments):
        """
        Populates the function arguments table with given arguments
        """
        self._arguments = arguments
        table = self.table
        table.setRowCount(len(arguments))
        for index, arg in enumerate(arguments):
            table.setItem(index, 0, QtWidgets.QTableWidgetItem(str(arg.ordinal)))
            table.setItem(index, 1, QtWidgets.QTableWidgetItem(str(arg.location)))
            table.setItem(index, 2, QtWidgets.QTableWidgetItem(arg.type))
            table.setItem(index, 3, QtWidgets.QTableWidgetItem(str(arg.width)))
            table.setItem(index, 4, QtWidgets.QTableWidgetItem(arg.name))
            addr = arg.addr
            table.setItem(index, 5, QtWidgets.QTableWidgetItem(hex(addr) if addr is not None else ""))
            table.setItem(index, 6, QtWidgets.QTableWidgetItem(hex(arg.value)))

            # If argument value is a pointer, show the first few bytes of the referenced data
            if self._ctx.memory.is_mapped(arg.value):
                data = self._ctx.memory.read(arg.value, 24)
                table.setItem(index, 7, QtWidgets.QTableWidgetItem(str(data) + "..."))
            else:
                table.setItem(index, 7, QtWidgets.QTableWidgetItem(""))

    def populate(self, ctx, num_args=None):
        self._ctx = ctx
        try:
            func_sig = ctx.get_function_signature(num_args=num_args)
            if not func_sig:
                raise dragodis.NotExistError("No Operand.")
        except dragodis.NotExistError:
            self.signature.setText("Function call not detected.")
            self.table.setRowCount(0)  # TODO: This clears the column information.
            return
        except Exception as e:
            self.signature.setText(f"<font color=red>{e}</font>")
            self.table.setRowCount(0)
            return

        self.num_args.setValue(len(func_sig.arguments))
        self.signature.setText(
            f"{hex(func_sig.address)}: {func_sig.declaration}"
        )
        self._populate_table(func_sig.arguments)



"""
Script to install IDA plugin.
"""

import pathlib
import shutil
import sys

from .util import get_ida_user_dir


if __name__ == "__main__":
    ida_dir = get_ida_user_dir()
    if not ida_dir:
        print("Unable to find IDA user directory.")
        sys.exit(1)
    entry_point = pathlib.Path(__file__).parent / "rugosa.py"
    destination = ida_dir / "plugins" / "rugosa.py"
    destination.parent.mkdir(parents=True, exist_ok=True)
    shutil.copy(entry_point, destination)
    print(f"Added {destination}")



"""
Entrypoint for the Rugosa IDA Plugin.
"""

from rugosa.ida_plugin.plugin import RugosaPlugin


def PLUGIN_ENTRY():
    return RugosaPlugin()


if __name__ == "__main__":
    # If user is running this like a script directly start it up.
    plugin = PLUGIN_ENTRY()
    plugin.init()
    plugin.run()



"""
Extra utility functions for interfacing within a dragodis disassembly
"""

from typing import Union, List, Iterable, Tuple, Optional

import logging

import dragodis
from dragodis import Disassembler, ReferenceType, NotExistError, OperandType
from dragodis.interface import Function, Line, Symbol

logger = logging.getLogger(__name__)


def iter_imports(dis: Disassembler, module_name=None, api_names=None) -> Iterable[Tuple[int, str, str]]:
    """
    Iterate the thunk function wrappers for API imports.
    Yields the module name, function name, and reference to function.

    .. code_block:: python

        for ea, name, module_name in utils.iter_imports():
            print("{}.{} function at: 0x{:0x}".format(module_name, name, ea))

        for ea, name, _ in utils.iter_imports("KERNEL32"):
            print("KERNEL32.{} function at: {}".format(name, ea))

        for ea, name, module_name in utils.iter_imports(api_names=["GetProcAddress", "GetFileSize"]):
            print("{}.{} function at: {}".format(module_name, name, ea))

        for ea, _, _ in utils.iter_imports("KERNEL32", "GetProcAddress"):
            print("KERNEL32.GetProcAddress function at: {}".format(ea))

    NOTE: The same function name can be yield more than once if it
    appears in multiple modules or has multiple thunk wrappers.

    Name is the original import name and does not necessarily reflect the function name.
    e.g. "GetProcAddress", "GetProcAddress_0", and "GetProcAddress_1" will all be "GetProcAddress"

    :param dis: Dragodis disassembler
    :param module_name: Filter imports to a specified library.
    :param api_names: Filter imports to specific API name(s).
        Can be a string of a single name or list of names.

    :yield: (ea, api_name, module_name)
    """
    if isinstance(api_names, str):
        api_names = [api_names]

    for import_ in dis.imports:
        # Some disassemblers include the ".dll" while others do not.
        namespace = import_.namespace.lower().rstrip(".dll")
        if module_name and namespace != module_name.lower():
            continue
        if _match_name(api_names, import_.name):
            yield import_.address, import_.name, namespace


# TODO: This function needs to be reworked.
def iter_dynamic_functions(dis: Disassembler) -> Iterable[Tuple[int, str]]:
    """
    Iterates the dynamically resolved function signatures.

    :yield: (ea, name)
    """
    try:
        data_segment = dis.get_segment(".data")
    except NotExistError:
        return
    for line in data_segment.lines:
        try:
            value = line.value
        except NotImplementedError:
            # TODO: For now, ignore dynamic functions within structs.
            continue
        if isinstance(value, int):
            try:
                dis.get_function(value)
                yield line.address, line.name
            except NotExistError:
                continue


def _match_name(func_names: Optional[List[str]], name: Optional[str]) -> bool:
    if not func_names:
        return True
    if not name:
        return False
    return (
        name in func_names
        or any(func_name in name for func_name in func_names)
    )


def iter_functions(dis: Disassembler, func_names: Union[None, str, List[str]] = None) -> Iterable[Tuple[int, str]]:
    """
    Iterate all defined functions and yield their address and name.
    (This includes imported and dynamically generated functions)

    :param func_names: Filter based on specific function names.

    :yield: (ea, name)
    """
    if isinstance(func_names, str):
        func_names = [func_names]
    
    emitted = set()

    # Yield declared functions.
    for func in dis.functions():
        name = func.name
        if _match_name(func_names, name):
            if func.start not in emitted:
                emitted.add(func.start)
                yield func.start, name

    # Also yield from imported.
    for import_ in dis.imports:
        if _match_name(func_names, import_.name):
            if import_.address not in emitted:
                emitted.add(import_.address)
                yield import_.address, import_.name

    # Yield dynamically resolved functions.
    for ea, name in iter_dynamic_functions(dis):
        if _match_name(func_names, name):
            if ea not in emitted:
                emitted.add(ea)
                yield ea, name


def _iter_calls_to(dis: Disassembler, target: Union[Line, Symbol]) -> Iterable[int]:
    for ref in target.references_to:
        if ref.type == ReferenceType.code_call:
            yield ref.from_address
        elif ref.type == ReferenceType.data_read:
            # some imports without a thunk function will only have a read reference
            # circumventing this with an explicit call instruction check at the read reference
            inst = dis.get_instruction(ref.from_address)
            if inst.is_call:
                yield ref.from_address


def iter_calls_to(dis: Disassembler, addr: int) -> Iterable[int]:
    """
    Iterates the calls to the given address.

    :param dis: dragodis Disassembler object
    :param func_ea: Address of a function.
    :yields: Call instruction address
    """
    line = dis.get_line(addr)
    yield from _iter_calls_to(dis, line)


def iter_import_calls(dis: Disassembler, name: str) -> Iterable[int]:
    """
    Iterates the calls to a given import by name.

    :param dis: dragodis Disassembler object
    :param name:  name of import function
    :yields: Call instruction address
    """
    imp = dis.get_import(name)
    yield from _iter_calls_to(dis, imp)


def iter_import_callers(dis: Disassembler, name: str) -> Iterable[Function]:
    """
    Iterates Function objects that call the given import.

    :param dis: dragodis Disassembler object
    :param name: name of import function
    :yields: Call instruction address
    """
    cache = set()
    for call_addr in iter_import_calls(dis, name):
        try:
            func = dis.get_function(call_addr)
        except NotExistError:
            continue
        if func.name not in cache:
            yield func
            cache.add(func.name)


def iter_callers(dis: Disassembler, addr: int) -> Iterable[Function]:
    """
    Iterates Function objects that call the given address.

    :param addr: Address of a function.
    :return:
    """
    cache = set()
    for call_addr in iter_calls_to(dis, addr):
        try:
            func = dis.get_function(call_addr)
        except NotExistError:
            continue
        if func.name not in cache:
            yield func
            cache.add(func.name)


def get_import_addr(dis: Disassembler, api_name, module_name=None) -> Optional[int]:
    """
    Returns the first instance of a function that wraps the given API name.

    .. code_block:: python

        proc_func_ea = get_import_addr("GetProcAddress")

    :param dis: Dragodis disassembler
    :param api_name: Name of API
    :param module_name: Library of API

    :returns: Address of function start or None if not found.
    """
    for ea, _, _ in iter_imports(dis, module_name, api_name):
        return ea


def get_export_addr(dis: Disassembler, export_name) -> Optional[int]:
    """
    Return the location of an export by name

    :param dis: Dragodis disassembler
    :param export_name: Target export

    :return: Location of target export or None
    """
    for export in dis.exports:
        if export.name == export_name:
            return export.address


def get_function_addr(dis: Disassembler, func_name: str) -> Optional[int]:
    """
    Obtain a function in the list of functions for the application by name.
    Supports using API resolved names if necessary.

    :param dis: Dragodis disassembler
    :param func_name: Name of function to obtain

    :return: start_ea of function or None
    """
    for ea, _ in iter_functions(dis, func_name):
        return ea


RAX_FAM = ["rax", "eax", "ax", "ah", "al"]


def find_destination(dis: Disassembler, start, instruction_limit=None) -> Optional[int]:
    """
    Finds the destination address for returned eax register.

    :param dis: Dragodis disassembler
    :param int start: Starting address to start looking
    :param int instruction_limit: Limit the number of instructions to traverse before giving up.
        Defaults to searching until the end of the function.

    :return: destination address or None if address couldn't be found or is not a loaded address.
    """
    count = 0
    func = dis.get_function(start)
    for insn in func.instructions(start):
        count += 1
        if instruction_limit is not None and count > instruction_limit:
            return None

        if insn.mnemonic == "mov" and insn.operands[1].text.lower() in RAX_FAM:
            if insn.operands[0].type == OperandType.memory:
                return insn.operands[0].value
            else:
                return None
    return None
